/*
 * Copyright 2002 Farfpeuf/ AdetSoft. All rights reserved.
 * Use is subject to license terms.
 */
package fb.army.entities;
import fb.battlefield.Sprite;
import fb.util.D6;
/**
 * a summoned daemon
 * @author farfpeuf
 */
public class Daemon extends SummonedCreature implements FlyingEntity {
	/**
	 * characteristics line number in catalog
	 */
	public static final int DAEMON_LINENUMBER = 138;
	/**
	 * delegate used for flying related methods
	 */
	private FlyingEntityDelegate flyingDelegate;

	/**
	 * get number of frames in entity sprite
	 */
	public int getNumberOfFrames() {
		return 96;
	}
	/**
	 * @return resource key
	 */
	protected String getResourceKey() {
		return "dragon";
	}
	
	
	/**
	 * create entity from line read in catalog
	 * 
	 * @param ligne
	 *            catalog line
	 * @param linenumber
	 *            line number
	 */
	public Daemon(String ligne, int linenumber) {
		super(ligne, linenumber);
		flyingDelegate = new FlyingEntityDelegate(this);
	}
	
	
	/**
	 * can this entity do a land manoeuvre
	 * 
	 * @return true if it can
	 */
	public boolean canLand() {
		return flyingDelegate.canLand();
	}
	/**
	 * can this entity do a take off manoeuvre
	 * 
	 * @return true if it can
	 */
	public boolean canTakeOff() {
		return flyingDelegate.canTakeOff();
	}
	/**
	 * create entity sprite
	 * 
	 */
	public void createSprite() {
		super.createSprite();
		flyingDelegate.createsprite();
	}
	/**
	 * delete sprite to free memory
	 * 
	 */
    public void freeSprite() {
        super.freeSprite();
        flyingDelegate.freeSprite();
    }

	public void doLand() {
		flyingDelegate.doLand();
	}
	/**
	 * do take off manoeuvre
	 * 
	 */
	public void doTakeOff() {
		flyingDelegate.doTakeOff();
	}
	/**
	 * @return sprite used when flying
	 */
	public Sprite getFlyingSprite() {
		return flyingDelegate.getFlyingSprite();
	}
	/**
	 * @return shadow sprite used when flying
	 */
	public Sprite getShadowSprite() {
		return flyingDelegate.getShadowSprite();
	}
	/**
	 * is this entity currently flying
	 * 
	 * @return true if this entity currently flying
	 */
	public boolean isFlying() {
		return flyingDelegate.isFlying();
	}
	/**
	 * land manoeuvre
	 * 
	 */
	public void land() {
		flyingDelegate.land();
	}
	/**
	 * returns entity as a string to save it to a file
	 * 
	 * @return entity as a string
	 */
	public String saveString() {
		return super.saveString() + flyingDelegate.savestring();
	}
	/**
	 * load entity from a StringBuffer
	 * 
	 * @param buffy
	 *            the StringBuffer
	 */
	public void loadString(StringBuffer buffy) {
		super.loadString(buffy);
		flyingDelegate.loadString(buffy);
	}
	/**
	 * take off manoeuvre
	 * 
	 */
	public void takeOff() {
		flyingDelegate.takeOff();
	}
	/**
	 * add sequences to sprite
	 */
	public void addSpriteSequences(Sprite sprite) {
		sprite.addSpritsequence("stand0", 28, 28, 28, 28);
		sprite.addSpritsequence("stand1", 29, 29, 29, 29);
		sprite.addSpritsequence("stand2", 30, 30, 30, 30);
		sprite.addSpritsequence("stand3", 31, 31, 31, 31);
		sprite.addSpritsequence("stand4", 24, 24, 24, 24);
		sprite.addSpritsequence("stand5", 25, 25, 25, 25);
		sprite.addSpritsequence("stand6", 26, 26, 26, 26);
		sprite.addSpritsequence("stand7", 27, 27, 27, 27);
		sprite.addSpritsequence("fight0", 68, 76, 84, 92);
		sprite.addSpritsequence("fight1", 69, 77, 85, 93);
		sprite.addSpritsequence("fight2", 70, 78, 86, 94);
		sprite.addSpritsequence("fight3", 71, 79, 87, 95);
		sprite.addSpritsequence("fight4", 64, 72, 80, 88);
		sprite.addSpritsequence("fight5", 65, 73, 81, 89);
		sprite.addSpritsequence("fight6", 66, 74, 82, 90);
		sprite.addSpritsequence("fight7", 67, 75, 83, 91);
		sprite.addSpritsequence("walk0", 36, 44, 52, 60);
		sprite.addSpritsequence("walk1", 37, 45, 53, 61);
		sprite.addSpritsequence("walk2", 38, 46, 54, 62);
		sprite.addSpritsequence("walk3", 39, 47, 55, 63);
		sprite.addSpritsequence("walk4", 32, 40, 48, 56);
		sprite.addSpritsequence("walk5", 33, 41, 49, 57);
		sprite.addSpritsequence("walk6", 34, 42, 50, 58);
		sprite.addSpritsequence("walk7", 35, 43, 51, 59);
		sprite.addSpritsequence("die0", 4, 12, 20, 20);
		sprite.addSpritsequence("die1", 5, 13, 21, 21);
		sprite.addSpritsequence("die2", 6, 14, 22, 22);
		sprite.addSpritsequence("die3", 7, 15, 23, 23);
		sprite.addSpritsequence("die4", 0, 8, 16, 16);
		sprite.addSpritsequence("die5", 1, 9, 17, 17);
		sprite.addSpritsequence("die6", 2, 10, 18, 18);
		sprite.addSpritsequence("die7", 3, 11, 19, 19);
	}
	/**
	 * calculate entity cost
	 */
	public int getCost() {
		return D6.roll();
	}
	/**
	 * summon a daemon
	 * @param master summoner
	 */
	public static void summon(SpellCaster master) {
		SummonedCreature.summon(master, DAEMON_LINENUMBER);
	}
	/**
	 * clone this daemon
	 */
	public Object clone() {
		Daemon daemon = (Daemon) super.clone();
		daemon.flyingDelegate = (FlyingEntityDelegate) flyingDelegate
				.clone(daemon);
		return daemon;
	}
	/**
	 * calculate magic cost
	 */
	public int magicCost() {
		return D6.roll();
	}
	/**
	 * try to move entity to target entity
	 * 
	 * @param e
	 *            target entity
	 * @param ex
	 *            destination x coordinate
	 * @param ey
	 *            destination y coordinate
	 * @return true if a path was found
	 */
	public boolean moveTo(Entity e, int ex, int ey) {
		boolean result = super.moveTo(e, ex, ey);
		if (result) {
			flyingDelegate.takeOffLandTo(e);
		}
		return result;
	}
}