/*
 * Copyright 2002 Farfpeuf/ AdetSoft. All rights reserved.
 * Use is subject to license terms.
 */
package fb.army.entities;

import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.Shape;
import java.awt.geom.AffineTransform;
import java.awt.geom.Area;
import java.util.ArrayList;
import java.util.List;

import fb.battlefield.BattleFieldModel;
import fb.battlefieldelements.Sign;
import fb.battlefieldelements.Warrior;
import fb.main.FB;
import fb.main.Battle;
import fb.util.D100;
import fb.util.DX;
import fb.util.FBColors;
import fb.util.GeometryTools;
import fb.util.Resources;
import fb.util.StringTools;

/**
 * an individual character
 * 
 * @author farfpeuf
 * 
 */
public abstract class Individual extends Entity {
	/**
	 * warrior model
	 */
	protected Warrior model;

	/**
	 * is this individual dead
	 */
	private boolean isdead = false;

	// used to hold warrios life points during unit creation

	private Integer lifePointsHolder;

	/**
	 * additionnal sign to visually identify individual type
	 */

	private Sign iconSign;

	/**
	 * create inidividual from line read in catalog
	 * 
	 * @param ligne
	 *            catalog line
	 * @param linenumber
	 *            line number
	 */
	public Individual(String ligne, int linenumber) {
		super(ligne, linenumber);
	}

	/**
	 * reset entity after end of battle
	 * 
	 * @param winner
	 *            did the entity army won the battle
	 */
	public void endBattle(boolean winner) {
		super.endBattle(winner);
		setState('n');
		lifePointsHolder = new Integer(model.getLifePoints());
		model = null;
		iconSign = null;
	}

	/**
	 * initialize entity at the beginning of a battle
	 * 
	 */
	public void initBattle() {
		super.initBattle();
		iconSign = new Sign(this);
		iconSign.setImage(Resources.getSign(getResourceKey()));
		signs.add(0, iconSign);
	}

	/**
	 * create warrior model
	 * 
	 */
	public void createModel() {
		model = new Warrior(this, (lifePointsHolder == null) ? initialPdv
				: lifePointsHolder.intValue());
		getModels().add(model);
		model.init();
		BattleFieldModel.getInstance().addWarrior(model);

	}

	/**
	 * @see fb.army.hierarchy.HierarchyNode#getNumberOfWarriors()
	 */
	public int getNumberOfWarriors() {
		return 1;
	}

	/**
	 * calculate base surface
	 * 
	 * @return base surface
	 */
	public float surface() {
		return getWidth() * getLength();
	}

	/**
	 * initialize life points
	 * 
	 * @param pdvs
	 */
	public void initLifePoints(int pdvs) {
		this.lifePointsHolder = new Integer(pdvs);
	}

	/**
	 * get individual total life points
	 * 
	 * @return individual total life points
	 */
	public int getTotalLifePoints() {
		return model.getLifePoints();
	}

	/**
	 * calculate individual cost
	 */
	public int getCost() {
		return price;
	}

	/**
	 * get front line width
	 * 
	 * @return front line width
	 */
	public int getFrontWidth() {
		return (int) (getWidth() * 8);
	}

	/**
	 * get max diagonal
	 * 
	 * @return maximum diagonal
	 */
	public int getMaxDiag() {
		float pythagore = getWidth() * getWidth() / 4f + getLength()
				* getLength() / 4f;
		return (int) Math.ceil(Math.sqrt(pythagore));
	}

	/**
	 * get Border
	 * 
	 * @return Border
	 */
	public Shape getBorder() {
		if (model == null) {
			return new Area();
		}
		return model.getBaseArea();
	}

	/**
	 * get x current coordinate
	 * 
	 * @return x current coordinate
	 */
	public int getX() {
		if (model != null) {
			return model.getXd();
		} else {
			return 0;
		}
	}

	/**
	 * get y current coordinate
	 * 
	 * @return y current coordinate
	 */
	public int getY() {
		if (model != null) {
			return model.getYd();
		} else {
			return 0;
		}
	}

	/**
	 * get Angle
	 * 
	 * @return Angle
	 */

	public double getAngle() {
		if (model != null) {
			return model.getAngle();
		} else {
			return 0;
		}
	}

	/**
	 * change angle
	 * 
	 * @param angle
	 *            new angle
	 */
	public void setAngle(double angle) {
		if (model != null) {
			model.setAngle(angle);
		}
	}

	/**
	 * get Contact Area
	 * 
	 * @return Contact Area
	 */
	public Area getContactArea() {
		return model.getContactArea();
	}

	/**
	 * get LeftMost Vertex
	 * 
	 * @return LeftMost Vertex
	 */
	public Point getLeftMostVertex() {
		return model.getA();
	}

	/**
	 * get RightMost Vertex
	 * 
	 * @return RightMost Vertex
	 */
	public Point getRightMostVertex() {
		return model.getB();
	}

	/**
	 * is this individual walking
	 * 
	 * @return true if individual is walking
	 */
	public boolean isWalking() {
		return model.iswalking();
	}

	/**
	 * is this individual dying
	 * 
	 * @return true if individual is dying
	 */
	public boolean isDying() {
		return model.getAction() == 'd';
	}

	/**
	 * stop individual action and make it stand
	 * 
	 */
	public void stand() {
		if (isActive() && model.getAction() != 'd' && model.getAction() != 'x') {
			model.setAction('n');
			model.attribseq();
		}
	}

	/**
	 * start fight animation
	 * 
	 */
	public void animFight() {
		model.animfight();
	}

	/**
	 * start threat animation
	 * 
	 */
	public void animThreat() {
		model.setAction('t');
		model.attribseq();
	}

	/**
	 * start death animation
	 * 
	 */
	public void animKill() {
		if (isdead) {
			model.setAction('d');
			model.setAnimationCounter(-DX.roll(10));
		}
	}

	/**
	 * calculate distance to another entity
	 * 
	 * @param e
	 *            the other entity
	 * @return distance to other entity
	 */
	public int distanceTo(Entity e) {
		return e.distanceFrom(model);
	}

	/**
	 * get front leftmost warrior
	 * 
	 * @return front leftmost warrior
	 */
	public Warrior leftModel() {
		return model;
	}

	/**
	 * get front rightmost warrior
	 * 
	 * @return front rightmost warrior
	 */
	public Warrior rightModel() {
		return model;
	}

	/**
	 * get translated line of movement
	 * 
	 * @param e
	 *            target entity
	 * @param tx
	 *            initial x coordinate
	 * @param ty
	 *            initial y coordinate
	 * @return line of movement length
	 */
	public int translatedLineOfMovementTo(Entity e, int tx, int ty) {
		return e.translatedLineOfMovementFrom(model, tx, ty);
	}

	/**
	 * get distance from a warrior
	 * 
	 * @param originmodel
	 *            the warrior
	 * @return distance from the warrior
	 */
	public int distanceFrom(Warrior originmodel) {
		int dx = originmodel.getX() - model.getX();
		int dy = originmodel.getY() - model.getY();
		return (int) java.lang.Math.sqrt(dx * dx + dy * dy);
	}

	/**
	 * get line of sight from a warrior
	 * 
	 * @param originmodel
	 *            the warrior
	 * @return line of sigh length
	 */
	public int lineOfSightFrom(Warrior originmodel) {
		Point losP0;
		Point losP1;
		losP0 = new Point(originmodel.getX(), originmodel.getY());
		losP1 = new Point(model.getX(), model.getY());

		if (clearLineOfSight(losP0, losP1, originmodel.getEntity())) {
			return distanceFrom(originmodel);
		} else {
			return Integer.MAX_VALUE;
		}
	}

	/**
	 * get translated line of movement from a warrior
	 * 
	 * @param originmodel
	 *            the warrior
	 * @param tx
	 *            initial x coordinate
	 * @param ty
	 *            initial y coordinate
	 * @return line of sigh length
	 */
	public int translatedLineOfMovementFrom(Warrior originmodel, int tx, int ty) {
		Point losP0;
		Point losP1;
		losP0 = new Point(originmodel.getX() + tx, originmodel.getY() + ty);
		losP1 = new Point(model.getX(), model.getY());
		if (clearLineOfMovement(losP0, losP1, originmodel.getEntity())) {
			return distanceFrom(originmodel);
		} else {
			return Integer.MAX_VALUE;
		}
	}

	/**
	 * fight
	 * 
	 */
	public void fight() {
		List weights, targets;
		Entity entity2;
		int index, nbopposing;
		weights = new ArrayList();
		targets = new ArrayList();
		nbopposing = model.getEnnemyContacts().size();
		for (int i = 0; i < nbopposing; i++) {
			entity2 = ((Warrior) (model.getEnnemyContacts().get(i)))
					.getEntity();
			index = targets.indexOf(entity2);
			if (index != -1) {
				double oldweight = ((Double) (weights.get(index)))
						.doubleValue();
				Double newweight = new Double(oldweight
						+ (double) (1 / nbopposing));
				weights.set(index, newweight);
			} else {
				targets.add(entity2);
				weights.add(new Double((double) (1 / nbopposing)));
			}
		} // for
		double reste, weight;
		int combatworth;
		reste = 0.0;
		for (int i = 0; i < targets.size(); i++) {
			weight = ((Double) (weights.get(i))).doubleValue();
			combatworth = (int) (weight * getWorth() + reste + 0.01);
			reste = weight * getWorth() - combatworth;
			combat(combatworth, (Entity) targets.get(i));
		}
	}

	/**
	 * combat another entity
	 * 
	 * @param combatworth
	 *            number of dice
	 * @param target
	 *            ennemy target
	 * @return number of wounds
	 */
	public int combat(int combatworth, Entity target) {
		// --hits
		int dicetohit = combatworth;
		int hits = 0, tohit = 50;
		if (level < target.level) {
			tohit += 15;
		} else if (level > target.level) {
			tohit -= 15;
		}
		if (!target.hasShield) {
			tohit -= 15;
		}
		for (int i = 0; i < dicetohit; i++) {
			int d100 = D100.roll();
			if (d100 >= D100.LUCKY) {
				hits++;
			} else if (d100 < D100.FUMBLE) {
			} else if (d100 >= tohit) {
				hits++;
			}
		}
		// --kills
		int kills = 0, tokill;

		tokill = rollToKill(target, weapon);
		tokill -= 15;
		tokill -= (getStrength() * 15);
		tokill += (target.getResistance() * 15);
		if (getState() == 'b') {
			tokill -= 15;
		}
		for (int i = 0; i < hits; i++) {
			int d100 = D100.roll();
			if (d100 >= D100.LUCKY) {
				kills++;
			} else if (d100 < D100.FUMBLE) {
			} else if (d100 >= tokill) {
				kills++;
			}
		}
		target.wound(kills);
		return kills;
	}

	/**
	 * calculate contacts
	 * 
	 */
	public void calculateContacts() {
		clearContacts();
		model.calculateContacts();
	}

	/**
	 * clear contacts
	 * 
	 */
	public void clearContacts() {
		friendlycontacts.clear();
		opposingContacts.clear();
		secondcontacts.clear();
		thirdcontacts.clear();
		model.clearcontacts();
	}

	/**
	 * calculate base area
	 * 
	 */
	public void calculateArea() {
		vertex = new Point[4];
		vertex[0] = model.getA();
		vertex[1] = model.getB();
		vertex[2] = model.getC();
		vertex[3] = model.getD();
		area = model.getBaseArea();
		if (isDeployed()) {
			calculateContacts();
		}
	}

	/**
	 * do move this individual to specified location
	 * 
	 * @param x
	 *            destination x coordinate
	 * @param y
	 *            destination y coordinate
	 * @param consumeMovement
	 *            must movement points be consumed
	 */
	public void doMoveTo(int nx, int ny, boolean consumeMovement) {
		if (consumeMovement) {
			consumeMovement((int) (0.99 + Math.sqrt((nx - model.getX())
					* (nx - model.getX()) + (ny - model.getY())
					* (ny - model.getY())) / 8));
		}
		model.moveTo(nx, ny);
		model.animMovement();
		calculateArea();
		int xd, yd;
		Sign sign;
		for (int i = 0; i < signs.size(); i++) {
			sign = (Sign) signs.get(i);
			xd = nx;
			yd = (int) ((i + 1) * getLength() * 8 + ny + 1);
			sign.moveto(xd, yd);
			sign.rotate(model.getAngle(), nx, ny);
		}
	}

	/**
	 * teleport this individual to its destination
	 * 
	 */
	public void teleport() {
		model.teleport();
		for (int i = 0; i < signs.size(); i++) {
			((Sign) signs.get(i)).teleport();
		}
	}

	/**
	 * do rotate this individual
	 * 
	 * @param theta
	 *            rotation angle
	 */
	public void doRotate(double theta) {
		model.rotate(theta);
		calculateArea();
		Sign sign;
		int xd, yd;
		for (int i = 0; i < signs.size(); i++) {
			sign = (Sign) signs.get(i);
			xd = model.getXd();
			yd = (int) ((i + 1) * getLength() * 8 + model.getYd() + 1);
			sign.moveto(xd, yd);
			sign.rotate(model.getAngle(), model.getXd(), model.getYd());
		}
	}

	/**
	 * left 45 manoeuvre
	 * 
	 */
	public void left45() {
		rotate(java.lang.Math.PI / 4);
	}

	/**
	 * right 45 manoeuvre
	 * 
	 */
	public void right45() {
		rotate(-java.lang.Math.PI / 4);
	}

	/**
	 * individual routs
	 */
	public void rout() {
		if (getState() != 'r' && getState() != 'a') {
			setState('r');
			model.rout();
			String message = Resources.getString(genre + "rout");
			;
			message = StringTools.replaceAll(message, "<entity full name>",
					fullName);
			FB.getInstance().message(message, getColor());
			if ((isLeader() || isMagic() || isHero() || isPriest() || isPaladin())) {
				getArmy().setMustTestMorale(true);
			}
		}
	}

	/**
	 * remove this individual after routing
	 * 
	 */
	public void commitRout() {

		command.removeActiveEntity(this);
		clearContacts();
		BattleFieldModel.getInstance().removeWarrior(model);
		lifePointsHolder = new Integer(model.getLifePoints());
		Battle.getInstance().getActiveEntities().remove(this);

		for (int i = 0; i < signs.size(); i++) {
			BattleFieldModel.getInstance().removeSign((Sign) signs.get(i));
		}
	}

	/**
	 * this individual is dead
	 */
	public void setDead(Warrior model) {
		isdead = true;
		if (((isLeader() || isMagic() || isHero() || isPriest() || isPaladin()) && type != 'u')) {
			getArmy().setMustTestMorale(true);
		}
	}

	/**
	 * wound this individual
	 * 
	 * @param wounds
	 *            number of wounds
	 * @return remaining wounds
	 */
	public int wound(int casualties) {
		int result = 0;
		model.wound(casualties);
		if (model.getLifePoints() <= 0) {
			result = 1;
		}
		return result;
	}

	/**
	 * remove dead warrior after killing
	 * 
	 */
	public void commitKill() {
		if (isdead) {
			// FB.debug("isdead " + isdead + "state " + getState());
			if (getState() != 'a') {
				setState('a');
				command.removeActiveEntity(this);
				String message = Resources.getString(genre + "dead");
				message = StringTools.replaceAll(message, "<entity full name>",
						fullName);
				FB.getInstance().message(message, getColor());
				clearContacts();
				// enlever la figurine
				BattleFieldModel.getInstance().removeWarrior(model);
				Battle.getInstance().getActiveEntities().remove(this);

				for (int i = 0; i < signs.size(); i++) {
					BattleFieldModel.getInstance().removeSign(
							(Sign) signs.get(i));
				}
			}
		}
	}

	/**
	 * bring individual back to life
	 * 
	 * @param priest
	 *            priest that resurected this individual
	 */
	public void resurection(Entity priest) {
		if (getSprite() == null) {
			createSprite();
		}
		initLifePoints(1);
		initBattle();
		create();
		setState('n');
		command.addActiveEntity(this);
		String message = Resources.getString(genre + "resurected");
		message = StringTools.replaceAll(message, "<entity full name>",
				fullName);
		FB.getInstance().message(message, getColor());

		Area deployArea = new Area(new Rectangle(priest.getX() - 100, priest
				.getY() - 100, priest.getX() + 100, priest.getY() + 100));
		deployArea.subtract(getMovementObstacles());
		boolean placed = false;
		int decX, decY, destX, destY;
		for (int i = 0; !placed && i < 20; i++) {
			decX = 10 * i / 2 * (-1 ^ (i % 2));
			destX = priest.getX() + decX;
			for (int j = 0; !placed && j < 20; j++) {
				decY = 10 * j / 2 * (-1 ^ (j % 2));
				destY = priest.getY() + decY;
				if (canDeployTo(destX, destY, deployArea)) {
					doDeployTo(destX, destY);
					placed = true;
				}
			}
		}
		;

		model.setAction('n');
		model.attribseq();
		Battle.getInstance().getActiveEntities().add(this);

	}

	/**
	 * paint movement range shape
	 * 
	 * @param g
	 *            graphics
	 * @param dx
	 *            x offset
	 * @param dy
	 *            y offset
	 * @param scale
	 *            view scale
	 */
	public void paintMovementRange(Graphics g, int dx, int dy, int scale) {

		destArea = getMovementRange();
		Area obstacles = getMovementObstacles();
		obstacles.subtract(this.area);
		destArea.subtract(obstacles);
		g.setColor(FBColors.secondary);
		Area scaledest = new Area(AffineTransform
				.getScaleInstance(scale, scale)
				.createTransformedShape(destArea));
		scaledest = new Area(AffineTransform.getTranslateInstance(dx, dy)
				.createTransformedShape(scaledest));
		((Graphics2D) g).draw(scaledest);
	}

	/**
	 * Sets the x.
	 * 
	 * @param x
	 *            The x to set
	 */
	public void setX(int x) {
		if (model != null) {
			model.setXd(x);
		}
	}

	/**
	 * Sets the y.
	 * 
	 * @param y
	 *            The y to set
	 */
	public void setY(int y) {
		if (model != null) {
			model.setYd(y);
		}
	}

	/**
	 * returns individual as a string to save it to a file
	 * 
	 * @return individual as a string
	 */
	public String saveString() {
		String result = super.saveString();
		if (model != null) {
			result += ("\t" + model.getLifePoints());
		} else if (lifePointsHolder != null) {
			result += ("\t" + lifePointsHolder.intValue());
		} else {
			result += ("\t" + initialPdv);
		}
		return result;
	}

	/**
	 * load individual from a StringBuffer
	 * 
	 * @param buffy
	 *            the StringBuffer
	 */
	public void loadString(StringBuffer buffy) {
		super.loadString(buffy);
		lifePointsHolder = new Integer(StringTools.readBufferInt(buffy));
	}

	/**
	 * get unit total worth
	 * 
	 * @return unit total worth
	 */
	public int totalWorth() {
		return getWorth();
	}

	/**
	 * shrink this unit
	 * 
	 */
	public void shrink() {
		super.shrink();
		model.resetAreas();
		calculateArea();
	}

	/**
	 * get initial life points
	 * 
	 * @return initial life points
	 */
	public int getInitialTotalLifePoints() {
		return initialPdv;
	}

	/**
	 * heal the individual
	 * 
	 * @param lifePoints
	 *            added life points
	 */
	public void heal(int healedPdvs) {
		model.heal(healedPdvs);
	}

	/**
	 * rotate individual to face another entity
	 * 
	 * @param e
	 *            the other entity
	 */
	public void rotateTo(Entity e) {
		int angleto = GeometryTools.to8Base(angleTo(e) - getAngle());
		if (angleto <= 4) {
			for (int i = 0; i < angleto; i++) {
				left45();
			}
		} else {
			for (int i = 8; i > angleto; i--) {
				right45();
			}
		}
	}

	/**
	 * number of warriors
	 * 
	 * @return number of warriors
	 */
	public int numberOfModels() {
		return 1;
	}

	/**
	 * @return the model
	 */
	public Warrior getModel() {
		return model;
	}

	/**
	 * @return true if individual is dead
	 */
	public boolean isDead() {
		return isdead;
	}

}