/*
 * Copyright 2002 Farfpeuf/ AdetSoft. All rights reserved.
 * Use is subject to license terms.
 */
package fb.artificialintelligence;

import java.util.Iterator;
import java.util.List;

import fb.util.D6;

/**
 * AI thread that chooses the best element among a set of availables elements
 * 
 * @author farfpeuf
 * 
 */
public class WeightChooser extends Thread {

	/**
	 * the current best element found
	 */
	protected Object chosenElement;

	/**
	 * the list of available elements among which the algorythm must find the
	 * best
	 */
	protected List availableElements;

	/**
	 * the number of d6 used for the random factor (the random factor is used to
	 * avoid predictability)
	 */
	private int nbD6;

	/**
	 * the weight of the best current element
	 */
	protected int maxWeight;

	/**
	 * true if the thread must be stopped at next iteration
	 */
	private boolean stop = false;

	/**
	 * this thread is stopped after a determined number of milliseconds The best
	 * element found (if any) is then available with getChosenElement method
	 */
	public void safeStop() {
		stop = true;
		interrupt();
	}

	/**
	 * create a WeightChooser to find the best element among a set of availables
	 * elements
	 * 
	 * @param availableElements
	 *            availables elements
	 * @param nbD6
	 *            numbers of d6 to roll to get the random factor
	 * @param minWeight
	 *            the minimum weight to have to be a good element (if none of
	 *            the availables elements obtain more than this weight, no
	 *            element will be selected)
	 */
	public WeightChooser(List availableElements, int nbD6, int minWeight) {
		super();
		setName("WeightChooser");
		this.availableElements = availableElements;
		this.nbD6 = nbD6;
		this.maxWeight = minWeight;
	}

	/**
	 * thread loop to find the best element
	 */
	public void run() {

		int weight;
		Object element;

		Iterator iter;

		iter = availableElements.iterator();
		while (iter.hasNext() && !stop && !interrupted()) {
			element = iter.next();
			weight = getWeight(element);
			if (weight > maxWeight) {
				maxWeight = weight;
				chosenElement = element;
			}
		}

		synchronized (this) {
			notify();
		}
	}

	/**
	 * calculate the weight for a specified element. the element with the
	 * biggest weight will be the choosen one. This method shall be redefined in
	 * sub-classes to add bonuses according to criterias that determine why an
	 * element is better than another
	 * 
	 * @param element
	 *            the element for which the weight has to be calculated
	 * @return
	 */
	protected int getWeight(Object element) {
		int weight = 0;
		// a random factor is used to avoid predictability
		for (int i = 0; i < nbD6; i++) {
			weight = weight + D6.roll();
		}
		return weight;
	}

	/**
	 * returns the best element found by the algorythm (if any)
	 * 
	 * @return the chosen element
	 */
	public Object getChosenElement() {
		return this.chosenElement;
	}

}