/*
 * Copyright 2002 Farfpeuf/ AdetSoft. All rights reserved.
 * Use is subject to license terms.
 */
package fb.battlefield;

import java.awt.AlphaComposite;
import java.awt.Color;
import java.awt.Component;
import java.awt.Composite;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Image;
import java.awt.MediaTracker;
import java.awt.Rectangle;
import java.awt.Shape;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import java.awt.font.FontRenderContext;
import java.awt.geom.AffineTransform;
import java.awt.geom.Area;
import java.awt.geom.Rectangle2D;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.util.Calendar;
import java.util.Iterator;
import java.util.List;

import javax.swing.RepaintManager;
import javax.swing.SwingUtilities;

import fb.army.entities.Entity;
import fb.army.entities.Shootingunit;
import fb.battlefieldelements.Element;
import fb.battlefieldelements.Figure;
import fb.battlefieldelements.MagicEffect;
import fb.battlefieldelements.Projectile;
import fb.battlefieldelements.Sign;
import fb.battlefieldelements.VerticalSet;
import fb.battlefieldelements.Warrior;
import fb.gui.frames.BattleFrame;
import fb.gui.panels.GamePanel;
import fb.main.FB;
import fb.main.Battle;
import fb.phases.BinaryPhase;
import fb.phases.DeploymentPhase;
import fb.phases.MovementPhase;
import fb.phases.PlayerMagicPhase;
import fb.phases.ShootingPhase;
import fb.util.FBColors;
import fb.util.Resources;

/**
 * the panel where the battlefield is drawn
 * 
 * @author farfpeuf
 * 
 */
public class BattleFieldPanel extends GamePanel implements
		BattleFieldModelListener, MouseMotionListener, MouseListener,
		PropertyChangeListener {

	// is mouse cursor over this panel ?
	private boolean mouseOver = true;

	// map image
	private Image mapImage;

	// current binary phase holder
	private BinaryPhase phase;

	// map width end lenght (unit is human warrior step)
	private int mapwidth, maplength;

	// shall map be painted
	private boolean paintMap = true;

	// shall light be painted (black filter that varry from day to night)
	private boolean paintLight = false;

	// shall vertical set be transparent
	private boolean transparentVSet = false;

	// scale (zoom factor)
	private int scale = 2;

	// current mousePosition holder
	private int mouseX, mouseY;

	// shall performance (number of milliseconds to draw a frame) be displayed
	private boolean displayPerformance = false;

	// animated cursor counter
	private int animatedCursorCounter = 0;

	/**
	 * create a new battlefield panel
	 * 
	 * 
	 */
	public BattleFieldPanel() {
		addMouseMotionListener(this);
		addMouseListener(this);
		RepaintManager.currentManager(this).setDoubleBufferingEnabled(true);
		BattleFieldModel.getInstance().addBFModelListener(this);
	}

	/**
	 * initialize panel
	 * 
	 * @param mapName
	 *            name of game map
	 * @param mapwidth
	 *            map width
	 * @param mapLength
	 *            map length
	 * @param flipLR
	 *            shall map be flipped from left to right
	 * @param flipUD
	 *            shall map be flipped from up to down
	 */
	public void init(String mapName, int mapwidth, int mapLength,
			boolean flipLR, boolean flipUD) {
		dx = 0;
		dy = 0;
		this.mapwidth = mapwidth;
		this.maplength = mapLength;
		mapImage = Resources.getMap(MapModel.getInstance().getImageName(),
				flipLR, flipUD);
		MediaTracker tracker = new MediaTracker(this);
		tracker.addImage(mapImage, 0);
		try {
			tracker.waitForAll();
		} catch (InterruptedException e) {
			e.printStackTrace();
		}
		if (!Battle.getInstance().isAutoDeploy()) {
			setScale(1);
		}

	}

	/**
	 * repaint the borders of a list of entities
	 * 
	 * @param entities
	 *            the list of entities
	 */
	public static void repaintEntitiesBorders(List entities) {
		if (entities != null) {

			for (int i = 0; i < entities.size(); i++) {
				repaintBorder((Entity) entities.get(i));
			}
		}
	}

	/**
	 * repaint the border of an entity
	 * @param entity the entity
	 */
	public static void repaintBorder(Entity entity) {
		if (entity != null) {
			Rectangle bounds = entity.getBorder().getBounds();
			refreshIt(bounds);
		}
	}

	/**
	 * action to perform when a property change event is fired
	 */
	public void propertyChange(PropertyChangeEvent evt) {
		Element element = (Element) evt.getSource();
		if (evt.getPropertyName().equals(Element.IMAGE)) {
			refreshIt(element.getBounds());
		} else if (evt.getPropertyName().equals(Element.BOUNDS)) {
			refreshIt((Rectangle) evt.getOldValue());
			refreshIt((Rectangle) evt.getNewValue());
		}
	}

	/**
	 * center view on specified point
	 * @param xGame x game coordinate
	 * @param yGame y game coordinate
	 */
	public void centerOn(int xGame, int yGame) {
		int bfpointx = xGame * scale + dx;
		int bfpointy = yGame * scale + dy;
		int bfcx = getWidth() / 2;
		int bfcy = getHeight() / 2;
		int xminus = bfpointx - bfcx;
		int yminus = bfpointy - bfcy;
		dx -= xminus;
		dy -= yminus;
		refreshAll();
	}

	/**
	 * center the view on the middle of the battlefield
	 *
	 */
	public void centerOnCenter() {
		centerOn(mapwidth / 2, maplength / 2);
	}

	/**
	 * refresh a rectangular section of the panel
	 * @param r the section to refresh
	 */
	public static synchronized void refreshIt(Rectangle r) {
		if (BattleFrame.getInstance().getBattlefieldPanel() != null) {
			BattleFrame.getInstance().getBattlefieldPanel().refresh(r);
		}
	}

	/**
	 * refresh a rectangular section of the panel
	 * @param r the section to refresh (game coordinate)
	 */
	public synchronized void refresh(Rectangle r) {
		int rx = r.x * scale + dx - 1;
		int ry = r.y * scale + dy - 1;
		int rw = r.width * scale + 2;
		int rh = r.height * scale + 2;

		super.refresh(new Rectangle(rx, ry, rw, rh));
	}

	//animate the elements at next time tic?
	private boolean animate;

	/**
	 * actions performed every timer tic
	 * 
	 */
	public void timeTic() {

		if (Battle.getLoop() != null
				&& Battle.getLoop().currentSynchronizedPhase instanceof BinaryPhase) {
			phase = (BinaryPhase) Battle.getLoop().currentSynchronizedPhase;
		} else {
			phase = null;
		}
		animate = !animate;
		if (animate) {
			Figure model;

			BattleFieldModel bfModel = BattleFieldModel.getInstance();

			for (int i = 0; i < bfModel.getFigures().size(); i++) {
				model = (Figure) bfModel.getFigures().get(i);
				model.animate();
			}
			Sign sign;
			for (int i = 0; i < bfModel.getSigns().size(); i++) {
				sign = (Sign) bfModel.getSigns().get(i);
				sign.animate();
			}
			MagicEffect effect;
			for (int i = 0; i < bfModel.getMagics().size(); i++) {
				effect = (MagicEffect) bfModel.getMagics().get(i);
				effect.animate();
			}
			Projectile projectile;
			for (int i = 0; i < bfModel.getProjectiles().size(); i++) {
				projectile = (Projectile) bfModel.getProjectiles().get(i);
				projectile.animate();
			}

		}

		super.timeTic();

	}

	/**
	 * stop animation
	 */
	public void stop() {
		super.stop();
		BattleFieldModel.getInstance().clear();
	}

	/** draw the battlefield
	 * 
	 */

	protected void draw(Graphics g) {

		long debut = Calendar.getInstance().getTimeInMillis();
		// peindre le fond
		if (paintMap) {
			drawMap(g);
		}
		if (phase != null && phase instanceof DeploymentPhase) {
			drawDeployement(g);
		} else {
			drawSigns(g);
			drawShapes(g);
			drawElements(g);
			drawArrowsAndEffects(g);
			if (paintLight) {
				drawLight(g);
			}
			if (displayPerformance) {
				long time = Calendar.getInstance().getTimeInMillis() - debut;
				drawPerformance(g, time);
			}
			// drawClip(g);
		}
	}

	/**
	 * draw performance (number of milliseconds to draw a frame and memory usage)
	 * @param g graphics
	 * @param time number of milliseconds to draw previous frame
	 */
	private void drawPerformance(Graphics g, long time) {
		g.setColor(FBColors.lightIvory);
		StringBuffer buff = new StringBuffer();
		buff.append("Milliseconds to draw frame : ");
		buff.append(time);
		buff.append(" memory in use : ");
		buff.append(Runtime.getRuntime().totalMemory()
				- Runtime.getRuntime().freeMemory());
		g.drawString(buff.toString(), 5, 20);
	}

	/**
	 * Draw deployment phase
	 * @param g graphics
	 */
	private void drawDeployement(Graphics g) {
		if (phase.getActor() != null) {
			// draw deployement area
			g.setColor(FBColors.secondary);

			Area scaledest = new Area(AffineTransform.getScaleInstance(scale,
					scale).createTransformedShape(
					((DeploymentPhase) phase).getDeployArea()));
			scaledest = new Area(AffineTransform.getTranslateInstance(dx, dy)
					.createTransformedShape(scaledest));

			((Graphics2D) g).draw(scaledest);

			if (mouseOver) {
				phase.getActor().paintDeployment(g, dx, dy, scale,
						((DeploymentPhase) phase).getDeployArea(), mouseX, mouseY);
			}
		}
		BattleFieldModel bfModel = BattleFieldModel.getInstance();
		bfModel.sort();
		Element element;
		for (int i = 0; i < bfModel.getElements().size(); i++) {
			element = (Element) bfModel.getElements().get(i);
			if (!(element instanceof Warrior)
					|| ((Warrior) element).getEntity().isDeployed()) {
				element.paintImage(g, dx, dy, scale);
			}
		}
	}

	/**
	 * draw game map
	 * @param g graphics
	 */
	private void drawMap(Graphics g) {

		g
				.drawImage(mapImage, dx, dy, mapwidth * scale, maplength
						* scale, null);

	}

	/**
	 * draw sign elements
	 * @param g graphics
	 */
	private void drawSigns(Graphics g) {
		BattleFieldModel bfModel = BattleFieldModel.getInstance();
		for (int i = 0; i < bfModel.getSigns().size(); i++) {
			((Sign) bfModel.getSigns().get(i)).paintImage(g, dx, dy, scale);
		}
	}

	/**
	 * draw geometric shapes
	 * @param g graphics
	 */
	private void drawShapes(Graphics g) {
		if (phase != null && phase.getCurrentArmy().isHandled()
				&& !phase.getCurrentArmy().isAi()) {
			if (phase.isActorSelection()) {
				g.setColor(FBColors.primary);
				for (int i = 0; i < phase.getPotentialActors().size(); i++) {
					((Entity) phase.getPotentialActors().get(i)).paintBorder(
							g, dx, dy, scale);
				}
			} else {
				g.setColor(FBColors.errorColor);
				for (int i = 0; i < phase.getPotentialTargets().size(); i++) {
					((Entity) phase.getPotentialTargets().get(i)).paintBorder(
							g, dx, dy, scale);
				}
				if (phase.getActor() != null) {
					g.setColor(FBColors.secondary);
					Entity actor = phase.getActor();
					actor.paintBorder(g, dx, dy, scale);
					if ((phase instanceof MovementPhase) && actor.isMoveAble()
							&& !actor.isWalking()) {
						actor.paintMovementRange(g, dx, dy, scale);
						if (mouseOver) {

							int okMove = ((MovementPhase) phase).getOkMove();
							actor.paintMovement(g, dx, dy, scale, mouseX,
									mouseY, okMove);
						}
					} else if ((phase instanceof ShootingPhase)) {
						((Shootingunit) actor).paintShootingRange(g, dx, dy,
								scale);
					}
				}
			}
		}
	}

	/**
	 * draw battlefield elements
	 * @param g graphics
	 */
	private void drawElements(Graphics g) {

		BattleFieldModel bfModel = BattleFieldModel.getInstance();
		bfModel.sort();
		Element element;
		for (int i = 0; i < bfModel.getElements().size(); i++) {
			element = (Element) bfModel.getElements().get(i);
			if (!(element instanceof VerticalSet) || !isTransparentVSet()) {
				element.paintImage(g, dx, dy, scale);
			}
		}
	}

	/**
	 * draw projectiles and magic effects
	 * @param g graphics
	 */
	private void drawArrowsAndEffects(Graphics g) {
		BattleFieldModel bfModel = BattleFieldModel.getInstance();
		for (int i = 0; i < bfModel.getProjectiles().size(); i++) {
			((Projectile) bfModel.getProjectiles().get(i)).paintImage(g, dx,
					dy, scale);
		}
		for (int i = 0; i < bfModel.getMagics().size(); i++) {
			((MagicEffect) bfModel.getMagics().get(i)).paintImage(g, dx, dy,
					scale);
		}
	}

	/**
	 * draw painting clip (used for debug)
	 * @param g graphics
	 */
	private void drawClip(Graphics g) {
		g.setColor(Color.white);
		Shape clip = g.getClip();
		((Graphics2D) g).draw(clip);
	}

	/**
	 * draw light filter (darker at night)
	 * @param g graphics
	 */
	private void drawLight(Graphics g) {
		Composite old_compo = ((Graphics2D) g).getComposite();
		float luminosity = 0f;
		int time = Battle.getInstance().getTime();
		if (time == 2 || time == 4) {
			luminosity = 0.1f;
		} else if (time == 1 || time == 5) {
			luminosity = 0.2f;
		} else if (time == 6 || time == 10) {
			luminosity = 0.3f;
		} else if (time == 7 || time == 9) {
			luminosity = 0.4f;
		} else if (time == 8) {
			luminosity = 0.5f;
		}
		((Graphics2D) g).setComposite(AlphaComposite.getInstance(
				AlphaComposite.SRC_OVER, luminosity));
		g.setColor(FBColors.background);
		g.fillRect(0, 0, getWidth(), getHeight());
		((Graphics2D) g).setComposite(old_compo);
	}

	/**
	 * switch transparentVSet value
	 *
	 */
	public void switchTransparentVSet() {
		transparentVSet = !transparentVSet;
		refreshAll();
	}

	/**
	 * @return transparentVSet value
	 */
	public boolean isTransparentVSet() {
		return transparentVSet;
	}

	/**
	 * area used to display performance
	 */
	private Rectangle perfoRectangle;

	/**
	 * switch displayPerformance value
	 *
	 */
	public void switchDisplayPerformance() {
		if (perfoRectangle == null) {
			Graphics2D gg = (Graphics2D) getGraphics();
			FontRenderContext frc = gg.getFontRenderContext();
			String perfo = "Milliseconds to draw frame : 8888 memory in use : 888888888";
			Rectangle2D rfl = gg.getFont().getStringBounds(perfo, frc);
			perfoRectangle = new Rectangle((int) Math.abs(rfl.getX()) - 1,
					(int) Math.abs(rfl.getY()) - 1, (int) Math.abs(rfl
							.getWidth()) + 1,
					(int) Math.abs(rfl.getHeight()) + 1);
		}
		;
		displayPerformance = !displayPerformance;
		refreshIt(perfoRectangle);
	}

	/**
	 * switch paintMap value
	 *
	 */
	public void switchPaintMap() {
		paintMap = !paintMap;
		refreshAll();
	}

	/**Light value
	 *
	 */
	public void switchPaintLight() {
		paintLight = !paintLight;
		refreshAll();
	}

	/**
	 * get x offset
	 */
	public int getDx() {
		return dx;
	}

	/**
	 * set x offset
	 */
	public synchronized void setDx(int dx) {
		this.dx = dx;
	}

	/**
	 * set y offset
	 */
	public int getDy() {
		return dy;
	}

	/**
	 * set y offset
	 */
	public synchronized void setDy(int dy) {
		this.dy = dy;
	}

	/**
	 * get scale
	 * @return zoom factor
	 */
	public int getScale() {
		return scale;
	}

	/**
	 * set scale
	 * @param scale zoom factor
	 */
	public synchronized void setScale(int scale) {
		this.scale = scale;
		refreshAll();
	}

	public void mouseDragged(MouseEvent e) {
		// nothing to do
	}

	/**
	 * get entity where mouse was clicked
	 * @param e mouse event produced by mouse click
	 * @return the entity or null if none
	 */
	private Entity getEntity(MouseEvent e) {
		Iterator iter = Battle.getInstance().getActiveEntities().iterator();
		Entity current;
		int mousex, mousey;
		mousex = (mouseX - dx) / scale;
		mousey = (mouseY - dy) / scale;
		while (iter.hasNext()) {
			current = (Entity) iter.next();
			if (current.getArea().contains(mousex, mousey)) {
				return current;
			}
		}
		return null;
	}

	//is mouse cursor over an entity
	private Entity mouseOverEntity;

	/**
	 * mouse motion handling
	 */
	public void mouseMoved(MouseEvent e) {
		mouseX = e.getX();
		mouseY = e.getY();
		mouseOver = true;
		if (phase != null && phase.isCurrentArmyHandledPlayer()
				&& !phase.isActorSelection()) {
			if (phase instanceof PlayerMagicPhase) {
				Entity newMouseOverEntity = getEntity(e);
				if (newMouseOverEntity != mouseOverEntity) {
					mouseOverEntity = newMouseOverEntity;
					setToolTipText(mouseOverEntity == null ? null
							: mouseOverEntity.getToolTip(phase));
				}
			} else if (phase instanceof MovementPhase) {
				int destx = (mouseX - dx) / scale;
				int desty = (mouseY - dy) / scale;
				int okMove = ((MovementPhase) phase).setDestination(destx,
						desty);
				if (okMove == MovementPhase.MOVE_TO_PLACE) {
					FB.getInstance().setBattlefieldCursor(FB.DEFAULT_CURSOR);
				} else if (okMove == MovementPhase.MOVE_TO_TARGET
						&& phase.getTarget() != null) {
					FB.getInstance().setBattlefieldCursor(FB.TARGET_CURSOR);
				} else {
					FB.getInstance().setBattlefieldCursor(FB.NO_CURSOR);
				}
			} else if (phase instanceof DeploymentPhase) {
				int destx = (mouseX - dx) / scale;
				int desty = (mouseY - dy) / scale;

				((DeploymentPhase) phase).setDestination(destx, desty);

			}
		}
	}

	/**
	 * get map length
	 * @return map length in game unit
	 */
	public int getMaplength() {
		return maplength;
	}
	/**
	 * get map width
	 * @return map width in game unit
	 */
	public int getMapwidth() {
		return mapwidth;
	}

	/**
	 * is mouse over panel
	 * @return true if mouse is over this panel
	 */
	public boolean isMouseOver() {
		return mouseOver;
	}

	public void mouseClicked(MouseEvent e) {
	}

	public void mouseEntered(MouseEvent e) {
	}

	public void mouseExited(MouseEvent e) {
		mouseOver = false;
	}

	public void mousePressed(MouseEvent e) {
	}

	public void mouseReleased(MouseEvent e) {
	}

	/**
	 * handle view motion and cursor image when mouse is on the edge of the panel
	 */
	protected void testEdgeMove() {
		if (isMouseOnEdges()) {
			boolean clipChanged = false;
			boolean n = false, o = false, s = false, e = false;
			int mouseX = getMouseCoord().x;
			int mouseY = getMouseCoord().y;

			Component root = SwingUtilities.getRoot(this);
			if (root != null) {
				if (mouseX <= 5) {
					o = true;
					setDx(getDx() + 30);
					clipChanged = true;
				} else if (mouseX >= root.getWidth() - 5) {
					e = true;
					setDx(getDx() - 30);
					clipChanged = true;
				}
				if (mouseY <= 5) {
					n = true;
					setDy(getDy() + 30);
					clipChanged = true;
				} else if (mouseY >= root.getHeight() - 5) {
					s = true;
					setDy(getDy() - 30);
					clipChanged = true;
				}
				if (clipChanged) {
					refreshAll();
				}
				if (n) {
					if (e) {
						FB.getInstance().setFrameCursor(FB.NORTH_EAST_CURSOR);
					} else if (o) {
						FB.getInstance().setFrameCursor(FB.NORTH_WEST_CURSOR);
					} else {
						FB.getInstance().setFrameCursor(FB.NORTH_CURSOR);
					}
				} else if (s) {
					if (e) {
						FB.getInstance().setFrameCursor(FB.SOUTH_EAST_CURSOR);
					} else if (o) {
						FB.getInstance().setFrameCursor(FB.SOUTH_WEST_CURSOR);
					} else {
						FB.getInstance().setFrameCursor(FB.SOUTH_CURSOR);
					}
				} else if (o) {
					FB.getInstance().setFrameCursor(FB.WEST_CURSOR);
				} else if (e) {
					FB.getInstance().setFrameCursor(FB.EAST_CURSOR);
				} else {
					FB.getInstance().setFrameCursor(FB.DEFAULT_CURSOR);
				}
			}
		} else if (FB.getInstance().isWaitCursorSet()) {
			if (animatedCursorCounter == 0) {
				FB.getInstance().setFrameCursor(FB.DRUM1);
			} else if (animatedCursorCounter == 1) {
				FB.getInstance().setFrameCursor(FB.DRUM2);
			} else if (animatedCursorCounter == 2) {
				FB.getInstance().setFrameCursor(FB.DRUM3);
			} else if (animatedCursorCounter == 3) {
				FB.getInstance().setFrameCursor(FB.DRUM4);
			}
			animatedCursorCounter = (animatedCursorCounter + 1) % 4;
		} else {
			if (mouseOver) {
				FB.getInstance().setFrameCursor(
						FB.getInstance().getBattlefieldCursor());
			} else {
				FB.getInstance().setFrameCursor(FB.DEFAULT_CURSOR);
			}
		}
	}

	/**
	 * overiden to center view when panel bounds are modified
	 */
	public void setBounds(int x, int y, int width, int height) {
		super.setBounds(x, y, width, height);
		centerOnCenter();
	}

	/**
	 * action performed when data model is changed
	 */
	public void bfChanged(BattleFieldModelEvent e) {
		switch (e.getType()) {
		case BattleFieldModelEvent.INSERT:
			e.getElement().addPropertyChangeListener(this);
			break;
		case BattleFieldModelEvent.DELETE:
			e.getElement().removePropertyChangeListener(this);
			break;
		}
		refreshIt(e.getElement().getBounds());
	}

	/**
	 * overridden to repaint minimap when big map must be repainted
	 */
	protected void testRepaint() {
		if (!dirtyRegions.isEmpty()) {
			BattleFrame.getInstance().getMinimap().repaint();
		}
		super.testRepaint();
	}

}