/*
 * Copyright 2002 Farfpeuf/ AdetSoft. All rights reserved.
 * Use is subject to license terms.
 */
package fb.phases;

import java.awt.Point;
import java.util.ArrayList;
import java.util.List;

import fb.army.entities.Entity;
import fb.army.hierarchy.Army;
import fb.artificialintelligence.AIConstants;
import fb.artificialintelligence.TargetFinder;
import fb.battlefield.BattleFieldPanel;
import fb.gui.frames.BattleFrame;
import fb.main.FB;
import fb.main.Battle;
import fb.network.Client;
import fb.network.MessagesConstants;
import fb.util.DX;
import fb.util.StringTools;

/**
 * a game phase where player chooses actors and targets
 * 
 * @author farfpeuf
 * 
 */
public abstract class BinaryPhase extends SynchronizedPhase {
	/**
	 * current actor and target
	 */
	private Entity actor, target;

	/**
	 * current selected entity
	 */
	protected Entity selected;

	/**
	 * current step is actor selection ?
	 */
	private boolean actorSelection = true;

	/**
	 * point clicked on the battlefield
	 */
	protected Point clickedPoint;

	/**
	 * available actors and targets
	 */
	private List potentialActors, potentialTargets;

	/**
	 * actors, targets and entities for which phase is finished
	 */
	protected List actors, targets, doneEntities;

	/**
	 * constructor initialize lists
	 */
	public BinaryPhase() {
		super();
		potentialActors = new ArrayList();
		potentialTargets = new ArrayList();
		actors = new ArrayList();
		targets = new ArrayList();
		doneEntities = new ArrayList();
	}

	/**
	 * initialize this phase
	 * 
	 */
	public void init() {
		super.init();
		initActorSelection();
	}

	/**
	 * @return true if current army is handled by this computer by a human
	 *         player
	 */
	public boolean isCurrentArmyHandledPlayer() {
		return armies[current].isHandled() && !armies[current].isAi();
	}

	/**
	 * wait for network action
	 */
	protected void waitNetAction() {
		FB.setWaitCursor();
		if (!stop) {
			if (Battle.getLoop().getPhaseTokens()[current]) {
				if (!finished()) {
					nextArmy();
				}
			}
		}
	}

	/**
	 * phase waiting for network action ?
	 */
	public boolean isWaitingForNetAction() {
		return (allHandledArmiesDone() && !finished())
				|| super.isWaitingForNetAction();
	}

	/**
	 * change current playing army
	 * 
	 */
	public void changeArmy() {
		int i = current + 1;
		if (i == armies.length) {
			i = 0;
		}
		while ((Battle.getLoop().getPhaseTokens()[i] || !armies[i].isHandled())
				&& !allHandledArmiesDone()) {
			i++;
			if (i == armies.length) {
				i = 0;
			}
		}
		if (armies[i].isHandled()) {
			current = i;
			// FB.debug("changeArmy " + armies[current].getPlayerName());
		}
		if (allHandledArmiesDone()) {
			// FB.debug("allHandledArmiesDone");
			Battle.getInstance().setArmy(null);
			waitNetEnd();
		} else {
			Battle.getInstance().setArmy(armies[current]);
		}

	}

	/**
	 * change current playing army and initialize actors selection
	 */
	public void nextArmy() {
		changeArmy();
		if (!allHandledArmiesDone()) {
			initActorSelection();
		}
	}

	/**
	 * wait for network end of phase
	 * 
	 */
	protected void waitNetEnd() {
		FB.setWaitCursor();
	}

	/**
	 * proceed to phase next step
	 */
	public void nextStep() {

		if (armies[current].isAi()) {
			handleAI();
		} else {
			waitForPlayerAction();
		}
	}

	/**
	 * handle human player action
	 * 
	 */
	public void handlePlayerAction() {
		Entity entity;
		boolean entityselected = false;
		List selectables;
		if (actorSelection) {
			selectables = getPotentialActors();
		} else {
			selectables = getPotentialTargets();
		}
		for (int i = 0; i < selectables.size(); i++) {
			entity = (Entity) (selectables.get(i));
			if (entity.getArea().contains(clickedPoint)) {
				entityselected = true;
				selected = entity;
			}
		}
		if (entityselected) {
			if (actorSelection) {
				setActor(selected);
				initTargetSelection();
			} else {
				doneEntities.add(actor);
				setTarget(selected);
				action();
				nextArmy();
			}
		} else {
			initActorSelection();
		}
	}

	/**
	 * handle computer player
	 * 
	 */
	public void handleAI() {

		if (actorSelection) {

			if (getPotentialActors().size() > 0) {

				setActor((Entity) getPotentialActors().get(
						DX.roll(getPotentialActors().size() - 1)));
				initTargetSelection();
			} else {
				if (!allHandledArmiesDone()) {
					nextArmy();
				}
			}
		} else {
			setTarget(getAITarget());
			doneEntities.add(actor);
			if (target != null) {
				action();
			}
			nextArmy();
		}
	}

	/**
	 * player Performed an Action
	 * 
	 */
	public void playerPerformedAction() {
		setWaitCursor();
		BattleFrame.getInstance().enableButtons(false);
		if (clickedPoint == null) { // nextphase
			setToken(current);
			if (!allHandledArmiesDone()) {
				nextArmy();
			}
		} else {
			handlePlayerAction();
		}
	}

	/**
	 * wait for player action
	 * 
	 */
	protected void waitForPlayerAction() {
		BattleFrame.getInstance().enableButtons(true);
		if (actorSelection) {
			setDefaultCursor();
		} else {
			setTargetCursor();
		}
		waitingForPlayerAction = true;
	}

	/**
	 * @return AI thread that handles choice of the best target among a set of
	 *         availables targets
	 */
	protected abstract TargetFinder getTargetFinder();

	/**
	 * calculate next target for computer player
	 * 
	 * @return the target
	 */
	public Entity getAITarget() {
		Entity target;
		TargetFinder targetFinder = getTargetFinder();
		synchronized (targetFinder) {
			targetFinder.start();
			try {
				targetFinder.wait(FB.getInstance().isBatch() ? 200
						: AIConstants.MAX_TARGET_SEARCHING_TIME);
			} catch (InterruptedException e) {
				FB.exception(e);
			}
		}
		target = targetFinder.getTarget();
		if (targetFinder.isAlive()) {
			TargetFinder moribund = targetFinder;
			targetFinder = null;
			moribund.safeStop();
		}
		return target;
	}

	/**
	 * initialize actor selection on the battlefield
	 * 
	 */
	public void initActorSelection() {

		setActorselection(true);
		if (armies[current].isHandled()
				&& !Battle.getLoop().getPhaseTokens()[current]) {
			calculatePotentialActors();
		}
	}

	/**
	 * initialize target selection on the battlefield
	 * 
	 */
	public void initTargetSelection() {
		setActorselection(false);
		calculatePotentialTargets();
	}

	/**
	 * calculate available actors
	 * 
	 */
	public void calculatePotentialActors() {
		Entity entity;
		clearPotentialActors();
		for (int i = 0; i < armies[current].getActiveEntities().size(); i++) {
			entity = (Entity) (armies[current].getActiveEntities().get(i));
			if (!doneEntities.contains(entity)) {
				if (isActor(entity)) {
					if (hasTargets(entity)) {
						addPotentialActor(entity);
					}
				}
			}
		} // for
		setActorselection(true);
		if (getPotentialActors().size() == 0) {
			setToken(current);
			if (!finished()) {
				nextArmy();
			}
		}
	}

	/**
	 * has this entity targets for this phase ?
	 * 
	 * @param actor
	 *            the phase actor
	 * @return true if this entity has targets for this phase ?
	 */
	public boolean hasTargets(Entity actor) {
		Entity entity;
		List entities, ennemies;
		ennemies = armies[current].getEnnemies();
		for (int i = 0; i < ennemies.size(); i++) {
			entities = ((Army) ennemies.get(i)).getActiveEntities();
			for (int j = 0; j < entities.size(); j++) {
				entity = (Entity) (entities.get(j));
				if (entity.isActive()) {
					if (isTarget(actor, entity)) {
						return true;
					}
				}
			}
		}
		return false;
	}

	/**
	 * calculate available targets
	 * 
	 */
	public void calculatePotentialTargets() {
		Entity entity;
		clearPotentialTargets();
		List ennemyEntities, ennemies;
		ennemies = armies[current].getEnnemies();
		for (int i = 0; i < ennemies.size(); i++) {
			ennemyEntities = ((Army) ennemies.get(i)).getActiveEntities();
			for (int j = 0; j < ennemyEntities.size(); j++) {
				entity = (Entity) (ennemyEntities.get(j));

				if (entity.isActive()) {
					if (isTarget(entity)) {
						addPotentialTarget(entity);
					}
				}
			}
		}
		setActorselection(false);
	}

	/**
	 * end the phase
	 */
	public void end() {
		calculateResults();
		if (FB.isNet() && !(this instanceof PlayerMagicPhase)
				&& !(this instanceof PlayerOrderPhase)) {

			Client.getInstance().synchronize();
		}
		Entity entity;
		for (int i = 0; i < Battle.getInstance().getActiveEntities().size(); i++) {
			entity = (Entity) Battle.getInstance().getActiveEntities().get(i);

			while (!stop && entity.isWalking()) {
				sleep(125);
			}
			if (entity.getState() == 'r') {
				entity.commitRout();
			}
		}

		clearPotentialActors();
		clearPotentialTargets();
		setActor(null);
		setTarget(null);
	}

	/**
	 * handler player click on the battlefield
	 * 
	 * @param p
	 *            clicked point
	 */
	public final void playerClicked(Point p) {
		boolean handleClick = false;
		synchronized (this) {
			if (waitingForPlayerAction && Battle.getLoop().interrupted == 0) {
				handleClick = true;
				waitingForPlayerAction = false;
			}
		}
		if (handleClick) {
			clickedPoint = p;

			playerPerformedAction();
			Battle.getLoop().resume();
		}

	}

	/**
	 * @return true if all the armies handled by this computer finished this
	 *         phase
	 */
	public boolean allHandledArmiesDone() {
		for (int i = 0; i < armies.length; i++) {
			if (armies[i].isHandled() && !Battle.getLoop().getPhaseTokens()[i]) {
				return false;
			}
		}
		return true;
	}

	/**
	 * validate new actor-target couple for this phase
	 * 
	 * @param actor
	 *            the actor
	 * @param target
	 *            the target
	 */
	public void addActorAndTarget(Entity actor, Entity target) {
		if (FB.isNet()) {
			Client.getInstance().sendAndWait(
					MessagesConstants.ACTORNTARGET + actor.getId() + "#"
							+ target.getId());
		} else {
			doAddActorAndTarget(actor, target);
		}
	}

	/**
	 * do validate new actor-target couple for this phase
	 * 
	 * @param actor
	 *            the actor
	 * @param target
	 *            the target
	 */
	public void doAddActorAndTarget(Entity actor, Entity target) {
		actors.add(actor);
		targets.add(target);
	}

	/**
	 * proceed to phase action
	 * 
	 */
	public void action() {
		addActorAndTarget(actor, target);
	}

	/**
	 * is this entity a potential target for specified actor
	 * 
	 * @param a
	 *            phase actor
	 * @param e
	 *            phase target
	 * @return true if this entity is a potential target for specified actor
	 */
	public abstract boolean isTarget(Entity a, Entity e);

	/**
	 * is this entity a potential target for current actor
	 * 
	 * @param e
	 *            phase target
	 * @return true if this entity is a potential target for current actor
	 */
	public abstract boolean isTarget(Entity e);

	/**
	 * is this entity a potential actor for this phase
	 * 
	 * @param a
	 *            phase actor
	 * @return true if this entity is a potential actor
	 */
	public abstract boolean isActor(Entity e);

	/**
	 * calculate phase results
	 * 
	 */
	public abstract void calculateResults();

	/**
	 * @return phase as a string to be saved to a file
	 */
	public String saveString() {
		String result = this.getClass().getName() + '\t' + current + '\t'
				+ (actor != null ? actor.getId() : -1) + '\t'
				+ (target != null ? target.getId() : -1) + '\t'
				+ (selected != null ? selected.getId() : -1) + '\t'
				+ (actorSelection ? 'y' : 'n') + '\t'
				+ (waitingForPlayerAction ? 'y' : 'n') + '\t'
				+ saveEntityList(getPotentialActors()) + '\t'
				+ saveEntityList(getPotentialTargets()) + '\t'
				+ saveEntityList(actors) + '\t' + saveEntityList(targets)
				+ '\t' + saveEntityList(doneEntities);
		return result;
	}

	/**
	 * load phase from string read in a file
	 * 
	 * @param buffy
	 *            the sting buffer
	 */
	public void loadString(StringBuffer buffy) {
		current = StringTools.readBufferInt(buffy);
		actor = Battle.getEntity(StringTools.readBufferInt(buffy));
		target = Battle.getEntity(StringTools.readBufferInt(buffy));
		selected = Battle.getEntity(StringTools.readBufferInt(buffy));
		actorSelection = (StringTools.readBufferChar(buffy) == 'y');
		waitingForPlayerAction = (StringTools.readBufferChar(buffy) == 'y');
		potentialActors = loadEntityList(buffy);
		potentialTargets = loadEntityList(buffy);
		actors = loadEntityList(buffy);
		targets = loadEntityList(buffy);
		doneEntities = loadEntityList(buffy);
	}

	/**
	 * load an entity list from string read in a file
	 * 
	 * @param buffy
	 *            the sting buffer
	 */
	public static ArrayList loadEntityList(StringBuffer buffy) {
		ArrayList result = new ArrayList();
		int size = StringTools.readBufferInt(buffy);
		int id;
		for (int i = 0; i < size; i++) {
			id = StringTools.readBufferInt(buffy);
			result.add(Battle.getEntity(id));
		}
		return result;
	}

	/**
	 * @return an entity list as a string to be saved to a file
	 */
	public static String saveEntityList(List list) {
		String result = "";
		result += list.size();
		int id;
		for (int i = 0; i < list.size(); i++) {
			id = ((Entity) list.get(i)).getId();
			result += "\t" + id;
		}
		return result;
	}

	/**
	 * set game default cursor
	 * 
	 */
	public void setDefaultCursor() {
		FB.getInstance().setBattlefieldCursor(FB.DEFAULT_CURSOR);
	}

	/**
	 * set target selection cursor
	 * 
	 */
	public void setTargetCursor() {
		FB.getInstance().setBattlefieldCursor(FB.TARGET_CURSOR);
	}

	/**
	 * set waiting cursor
	 * 
	 */
	public void setWaitCursor() {
		FB.setWaitCursor();
	}

	/**
	 * interrupt this phase
	 * 
	 */
	public void interrupt() {
		clearPotentialActors();
		clearPotentialTargets();
		setActor(null);
		setTarget(null);
	}

	/**
	 * resume phase
	 * 
	 */
	public void resume() {
		initActorSelection();
		BattleFrame.getInstance().setPhase(this);
		BattleFieldPanel.repaintEntitiesBorders(potentialActors);
		if (armies[current].isHandled() && !armies[current].isAi()) {
			BattleFrame.getInstance().enableButtons(true);
		}
		setDefaultCursor();
	}

	/**
	 * @return Returns the actor.
	 */
	public Entity getActor() {
		return actor;
	}

	/**
	 * @return Returns the actorSelection.
	 */
	public boolean isActorSelection() {
		return actorSelection;
	}

	/**
	 * set phase actor
	 * 
	 * @param actor
	 *            the actor
	 */
	public void setActor(Entity actor) {
		Entity oldActor = this.actor;
		this.actor = actor;
		if (!armies[current].isAi()) {
			BattleFieldPanel.repaintBorder(oldActor);
			BattleFieldPanel.repaintBorder(actor);
		}
	}

	/**
	 * switch actor selection mode
	 * 
	 * @param actorselection
	 *            actor selection or target selection ?
	 */
	public void setActorselection(boolean actorselection) {

		this.actorSelection = actorselection;
		if (!armies[current].isAi()) {
			if (actor != null) {
				BattleFieldPanel.repaintBorder(actor);
			}
			if (target != null) {
				BattleFieldPanel.repaintBorder(target);
			}
			BattleFieldPanel.repaintEntitiesBorders(potentialActors);
			BattleFieldPanel.repaintEntitiesBorders(potentialTargets);
		}
	}

	/**
	 * set phase target
	 * 
	 * @param actor
	 *            the target
	 */
	public void setTarget(Entity target) {
		Entity oldTarget = this.target;
		this.target = target;
		if (!armies[current].isAi()) {
			BattleFieldPanel.repaintBorder(oldTarget);
			BattleFieldPanel.repaintBorder(this.target);
		}
	}

	/**
	 * @return Returns the potentialActors.
	 */
	public List getPotentialActors() {
		return potentialActors;
	}

	/**
	 * @return Returns the potentialTargets.
	 */
	public List getPotentialTargets() {
		return potentialTargets;
	}

	/**
	 * @return Returns the target.
	 */
	public Entity getTarget() {
		return target;
	}

	/**
	 * clear Potential Targets
	 * 
	 */
	public void clearPotentialTargets() {

		BattleFieldPanel.repaintEntitiesBorders(potentialTargets);

		potentialTargets.clear();
	}

	/**
	 * add a potential target
	 * 
	 * @param potentialTarget
	 *            the potential target to add
	 */
	public void addPotentialTarget(Entity potentialTarget) {
		potentialTargets.add(potentialTarget);
		if (!armies[current].isAi()) {
			BattleFieldPanel.repaintBorder(potentialTarget);
		}
	}

	/**
	 * add potential targets
	 * 
	 * @param targetsToAdd
	 *            the potential targets
	 */
	public void addPotentialTargets(List targetsToAdd) {
		this.potentialTargets.addAll(targetsToAdd);
		if (!armies[current].isAi()) {
			BattleFieldPanel.repaintEntitiesBorders(targetsToAdd);
		}
	}

	/**
	 * remove a potential target
	 * 
	 * @param potentialTarget
	 *            the potential target to remove
	 */
	public void removePotentialTarget(Entity potentialTarget) {
		potentialTargets.remove(potentialTarget);
		if (!armies[current].isAi()) {
			BattleFieldPanel.repaintBorder(potentialTarget);
		}
	}

	/**
	 * clear Potential Actors
	 * 
	 */
	public void clearPotentialActors() {

		BattleFieldPanel.repaintEntitiesBorders(potentialActors);

		potentialActors.clear();
	}

	/**
	 * add a potential actor
	 * 
	 * @param potentialActor
	 *            the potential actor to add
	 */
	public void addPotentialActor(Entity potentialActor) {
		potentialActors.add(potentialActor);
		if (!armies[current].isAi()) {
			BattleFieldPanel.repaintBorder(potentialActor);
		}
	}

	/**
	 * remove a potential actor
	 * 
	 * @param potentialActor
	 *            the potential actor to remove
	 */
	public void removePotentialActor(Entity potentialActor) {
		potentialActors.remove(potentialActor);
		if (!armies[current].isAi()) {
			BattleFieldPanel.repaintBorder(potentialActor);
		}
	}
}