/*
 * Copyright 2002 Farfpeuf/ AdetSoft. All rights reserved.
 * Use is subject to license terms.
 */
package fb.phases;

import java.awt.Rectangle;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import fb.army.entities.Daemon;
import fb.army.entities.Dragon;
import fb.army.entities.Entity;
import fb.army.entities.Heroic;
import fb.army.entities.Leader;
import fb.army.entities.Priest;
import fb.army.entities.Shootingunit;
import fb.army.entities.SpellCaster;
import fb.army.entities.SummonedCreature;
import fb.army.entities.Unit;
import fb.army.hierarchy.Army;
import fb.army.hierarchy.Command;
import fb.army.hierarchy.MercenaryCommand;
import fb.artificialintelligence.TargetFinder;
import fb.battlefield.BattleFieldPanel;
import fb.gui.frames.BattleFrame;
import fb.main.FB;
import fb.main.Battle;
import fb.network.Client;
import fb.network.MessagesConstants;
import fb.sound.SoundPlayer;
import fb.util.DX;
import fb.util.NetworkDX;
import fb.util.StringTools;

public class MovementPhase extends BinaryPhase implements
		PropertyChangeListener {

	// movement types constants

	public static final int MOVE_TO_PLACE = 1;

	public static final int MOVE_TO_TARGET = 0;

	public static final int DO_NOT_MOVE = -1;

	/**
	 * phase name
	 */
	private static final String name = "movementphase";

	/**
	 * phase icon number
	 */
	private static final int iconNumber = 2;

	/**
	 * ended turns tokens
	 */
	private boolean endOfTurn[];

	/**
	 * is entity movement possible
	 */
	private int okMove;

	/**
	 * entity destination x and y coordinate
	 */
	private int destX, destY;

	/**
	 * current command index
	 */
	private int currentCommandIndex;

	/**
	 * current command
	 */
	private Command currentCommand;

	/**
	 * commands that finished to play this phase
	 */
	private List doneCommands;

	/**
	 * constructor initilize arrays and vectors
	 */
	public MovementPhase() {
		super();
		setWithPhasePanel(true);
		doneCommands = new ArrayList();
		endOfTurn = new boolean[Battle.getInstance().getArmies().length];
	}

	/**
	 * reset ended turns tokens
	 * 
	 */
	public void resetEOT() {
		synchronized (endOfTurn) {
			for (int i = 0; i < endOfTurn.length; i++) {
				endOfTurn[i] = false;
			}
		}
	}

	/**
	 * @return true if all armies ended their turn
	 */
	private boolean allEOT() {
		synchronized (endOfTurn) {
			for (int i = 0; i < endOfTurn.length; i++) {
				if (!endOfTurn[i]) {
					return false;
				}
			}
			return true;
		}
	}

	/**
	 * test if army ended its turn
	 * 
	 * @param armyId
	 *            the army id
	 * @return true if army ended its turn
	 */
	public boolean isEOT(int armyId) {
		synchronized (endOfTurn) {
			return endOfTurn[armyId];
		}
	}

	/**
	 * end army turn
	 * 
	 * @param armyId
	 *            the army id
	 */
	public synchronized void endOfTurn(int armyId) {
		synchronized (endOfTurn) {

			endOfTurn[armyId] = true;
			if (!armies[current].isHandled()) {
				if (armyId == current) {
					Battle.getLoop().resume();
				}
			}

		}
	}

	/**
	 * phase waiting for network action ?
	 */
	public boolean isWaitingForNetAction() {
		return !armies[current].isHandled() && !isEOT(current)
				&& !Battle.getLoop().getPhaseTokens()[current];
	}

	/**
	 * wait for network action
	 */
	protected void waitNetAction() {

		if (isEOT(current)) {
			if (!finished()) {
				nextArmy();
			}
		} else {
			super.waitNetAction();
		}
	}

	/**
	 * @return phase name
	 */
	public String getName() {
		return name;
	}

	/**
	 * @return phase icon number
	 */
	public int getIconNumber() {
		return iconNumber;
	}

	/**
	 * proceed to phase next step
	 */
	public void nextStep() {
		if (armies[current].isHandled()) {
			if (armies[current].isAi()) {
				handleAI();
			} else {
				waitForPlayerAction();
			}
		} else {
			waitNetAction();
		}

	}

	/**
	 * handle entity property change
	 */
	public void propertyChange(PropertyChangeEvent evt) {
		Entity source = (Entity) evt.getSource();
		int oldRemainingMove = ((Integer) evt.getOldValue()).intValue();
		Rectangle dirty = source.getMovementRangeBounds(oldRemainingMove);
		BattleFieldPanel.refreshIt(dirty);
	}

	/**
	 * switch actor selection mode
	 * 
	 * @param actorselection
	 *            actor selection or target selection ?
	 */
	public void setActorSelection(boolean actorselection) {
		Entity actor = getActor();
		if (actor != null) {
			BattleFieldPanel.refreshIt(actor.getMovementRangeBounds());
		}
		super.setActorselection(actorselection);
	}

	/**
	 * set phase actor
	 * 
	 * @param actor
	 *            the actor
	 */
	public void setActor(Entity actor) {
		if (getActor() != null) {
			getActor()
					.removePropertyChangeListener(Entity.REMAINING_MOVE, this);
			BattleFieldPanel.refreshIt(getActor().getMovementRangeBounds());
		}
		super.setActor(actor);
		if (armies[current].isHandled() && !armies[current].isAi()) {
			BattleFrame.getInstance().setMovementPanel(actor);
		}
		if (actor != null) {
			actor.addPropertyChangeListener(Entity.REMAINING_MOVE, this);
			BattleFieldPanel.refreshIt(actor.getMovementRangeBounds());
		}
	}

	/**
	 * set actor destination
	 * 
	 * @param destX
	 *            destination x coordinate
	 * @param destY
	 *            destination y coordinate
	 * @return type of movement
	 */
	public int setDestination(int destX, int destY) {
		int oldOkMove = okMove;
		int oldDestX = this.destX;
		int oldDestY = this.destY;
		this.destX = destX;
		this.destY = destY;
		Entity actor = getActor();
		if (actor != null) {
			if (actor.getDestArea() != null
					&& actor.getDestArea().contains(destX, destY)) { // over
				// empty
				// space
				if (actor.canMoveTo(destX, destY, (Entity) null)) {
					okMove = MOVE_TO_PLACE;
				} else {
					okMove = DO_NOT_MOVE;
				}
				setTarget(null);
			} else { // over target
				Entity entity, target = null;
				okMove = DO_NOT_MOVE;
				Iterator iter = Battle.getInstance().getActiveEntities()
						.iterator();
				while (iter.hasNext()) {
					entity = (Entity) iter.next();
					if (entity != actor
							&& entity.getArea().contains(destX, destY)) {
						target = entity;
						okMove = MOVE_TO_TARGET;
						break;
					}
				}

				setTarget(target);

			}
			if (oldOkMove != okMove || oldDestX != destX || oldDestY != destY) {
				BattleFieldPanel.refreshIt(getActor().getDestinationBounds());
			}
		}
		return okMove;
	}

	/**
	 * @return phase as a string to be saved to a file
	 */
	public String saveString() {
		String result = super.saveString();
		result += "\t" + currentCommandIndex;
		result += '\t' + saveCommandList(doneCommands);
		return result;
	}

	/**
	 * load phase from string read in a file
	 * 
	 * @param buffy
	 *            the sting buffer
	 */
	public void loadString(StringBuffer buffy) {
		super.loadString(buffy);
		currentCommandIndex = StringTools.readBufferInt(buffy);
		currentCommand = (Command) armies[current].getAllGroups().get(
				currentCommandIndex);
		doneCommands = loadCommandList(buffy);
		if (armies[current].isHandled() && !armies[current].isAi()) {
			BattleFrame.getInstance().setMovementPanel(getActor());
		}
	}

	/**
	 * load an command list from string read in a file
	 * 
	 * @param buffy
	 *            the sting buffer
	 */
	public static ArrayList loadCommandList(StringBuffer buffy) {
		ArrayList result = new ArrayList();
		int size = StringTools.readBufferInt(buffy);
		int armyId, groupId;
		for (int i = 0; i < size; i++) {
			armyId = StringTools.readBufferInt(buffy);
			groupId = StringTools.readBufferInt(buffy);
			result.add(Battle.getGroup(armyId, groupId));
		}
		return result;
	}

	/**
	 * @return an command list as a string to be saved to a file
	 */
	public static String saveCommandList(List list) {
		String result = "";
		result += list.size();
		Command group;
		for (int i = 0; i < list.size(); i++) {
			group = (Command) list.get(i);
			result += "\t" + group.getArmy().getGmId() + "\t"
					+ group.getArmIndex();
		}
		return result;
	}

	/**
	 * change current playing army and initialize actors selection
	 */
	public void nextArmy() {
		if (allEOT()) {
			resetEOT();
		}
		changeArmy();
		if (!finished()) {
			initActorSelection();
		}
	}

	/**
	 * change current playing army
	 * 
	 */
	public void changeArmy() {

		int i = current + 1;
		if (i == armies.length) {
			i = 0;
		}
		while (Battle.getLoop().getPhaseTokens()[i] && !finished()) {
			i++;
			if (i == armies.length) {
				i = 0;
			}
		}
		current = i;
		// FB.debug("changeArmy " + armies[current].getPlayerName());

		if (!finished()) {
			Battle.getInstance().setArmy(armies[current]);
			if (armies[current].isHandled()) {
				initCurrentCommand();
				if (hasCommandFinished(currentCommandIndex)) {
					changeCommand();
				}
			}

		} else {
			Battle.getInstance().setArmy(null);
		}

	}

	/**
	 * handle human player action
	 * 
	 */
	public void handlePlayerAction() {
		if (isActorSelection()) {
			Entity entity;
			boolean entityselected = false;
			for (int i = 0; i < getPotentialActors().size(); i++) {
				entity = (Entity) (getPotentialActors().get(i));
				if (entity.getArea().contains(clickedPoint)) {
					entityselected = true;
					selected = entity;
				}
			}
			if (entityselected) {
				setActor(selected);
				setActorselection(false);
			}
		} else {
			if (getActor().isMoveAble()
					&& (okMove == MOVE_TO_PLACE || okMove == MOVE_TO_TARGET)) {
				action();
			} else {
				initActorSelection();
			}
		}
	}

	/**
	 * player Performed an Action
	 * 
	 */
	public void playerPerformedAction() {
		setWaitCursor();
		BattleFrame.getInstance().enableButtons(false);
		if (clickedPoint == null) { // nextphase
			commandFinished();
			if (!finished()) {
				nextArmy();
			}
		} else {
			handlePlayerAction();
		}
	}

	/**
	 * current command finished its movements
	 * 
	 */
	private void commandFinished() {

		doneCommands.add(currentCommand);
		if (allCommandFinished()) {
			setToken(current);
		} else {
			if (FB.isNet()) {
				Client.getInstance().sendAndWait(
						MessagesConstants.ENDTURN + current);
			}
		}

	}

	/**
	 * handle computer player
	 * 
	 */
	public void handleAI() {

		if (isActorSelection()) {
			setActor((Entity) getPotentialActors().get(
					DX.roll(getPotentialActors().size() - 1)));
			setActorselection(false);
		} else {
			setTarget(getAITarget());
			if (getTarget() != null) {
				manoeuvreAI(getTarget());
				if (getActor().isMoveAble()) {
					iaAction();
				}
			}
			getActor().setRemainingmove(0);

			initActorSelection();

		}

	}

	/**
	 * handle computer player manoeuvres
	 * 
	 */
	public void manoeuvreAI(Entity target) {
		getActor().rotateTo(target);
		if (getActor().isUnit()) {
			Unit u = (Unit) getActor();
			if (u.getOpposingContacts().size() > 0) {
				if (u.canSurround()) {
					u.surround();
				}
				int canExpend = u.canExpend();
				if (canExpend > 0) {
					u.expend(canExpend);
				}
			} else {
				int canReduce = u.canReduce();
				if (canReduce > 0) {
					int idealFront = (int) Math.sqrt(u.getNumber()) + 1;
					if (idealFront > u.getNumber()) {
						idealFront = u.getNumber();
					}
					if (u.getFront() > idealFront) {
						if (u.getFront() - idealFront >= canReduce) {
							u.reduce(canReduce);
						} else {
							u.reduce(u.getFront() - idealFront);
						}
					}
				}
			}
		}
	}

	/**
	 * @return AI thread that handles choice of the best target among a set of
	 *         availables targets
	 */
	public TargetFinder getTargetFinder() {
		clearPotentialTargets();
		Iterator iter = getActor().getArmy().getEnnemies().iterator();
		Army ennemyArmy;
		while (iter.hasNext()) {
			ennemyArmy = (Army) iter.next();
			addPotentialTargets(ennemyArmy.getActiveEntities());
		}

		// FB.debug(getActor().getFullName());

		return new TargetFinder(getPotentialTargets(), 3, 15, getActor()) {

			protected int getWeight(Object element) {
				Entity potentialTarget = (Entity) element;
				int poid = super.getWeight(potentialTarget);
				if (entity.isEnnemy(potentialTarget)) {
					if (potentialTarget.suffersBadLight()) {
						poid += 1;
					}
					if (!entity.suffersBadLight()) {
						poid += 1;
					}
					if (entity.getCommand() instanceof MercenaryCommand) {
						poid += 1;
					}
					if (entity instanceof SpellCaster) {
						if (entity.stepsTo(potentialTarget) > entity
								.getMovement()
								+ potentialTarget.getMovement() * 2) {
							poid += 4;
						}

					}

					if (entity instanceof Shootingunit
							&& ((Shootingunit) entity).getProjectilesNumber() > 0) {
						if (potentialTarget.stepsTo(entity) > (potentialTarget
								.getRemainingMove() + entity.getRemainingMove())) {
							poid += 4;
							if (entity.stepsTo(potentialTarget) > entity
									.getMovement()
									+ potentialTarget.getMovement() * 2) {
								poid += 2;
							}

							if (entity instanceof Dragon) {
								poid += 3;
							}
						}
					} else {
						if (entity instanceof Heroic || entity instanceof Unit
								|| entity instanceof Daemon) {
							poid += 7;
							if (potentialTarget.getOpposingContacts().size() == 0) {
								poid += 2;
							}
							if (entity.canReach(potentialTarget,
									getMouvementObstacles(potentialTarget))) {
								poid += 2;
							}
							if (potentialTarget == getClosestTarget()) {
								poid += 2;
							}
							if (potentialTarget.isShootingUnit()) {
								poid += 1;
							}
						}
					}
				} else {
					if (entity instanceof Priest) {
						if (potentialTarget.isActive()
								&& potentialTarget.hasLostLifePoints()) {
							poid += 5;
							if (potentialTarget.getTotalLifePoints() == 1) {
								poid += 3;
							}
							if (potentialTarget instanceof Leader
									|| potentialTarget instanceof Heroic) {
								poid += 3;
							}
						}

					}
				}
				// FB.debug(potentialTarget.getFullName()+" - > "+poid);
				return poid;
			}
		};
	}

	/**
	 * proceed to computer player action
	 * 
	 */
	public void iaAction() {
		getActor().moveTo(getTarget());
		if (getActor().isUnit()) {
			SoundPlayer.getInstance().play("walk");
		} else {
			SoundPlayer.getInstance().play("indvdwalk");
		}
		getActor().clearContacts();
		getActor().calculateContacts();
	}

	/**
	 * calculate available actors
	 * 
	 */
	public void calculatePotentialActors() {
		Entity entity;
		clearPotentialActors();
		List activeEntities = currentCommand.getActiveEntities();

		for (int i = 0; i < activeEntities.size(); i++) {
			entity = (Entity) (activeEntities.get(i));
			if (isActor(entity)) {
				addPotentialActor(entity);
			}
		}
		setActorselection(true);
		if (getPotentialActors().size() == 0) {
			commandFinished();
			if (!finished()) {
				nextArmy();
			}

		}
	}

	/**
	 * is this entity a potential target for specified actor
	 * 
	 * @param a
	 *            phase actor
	 * @param e
	 *            phase target
	 * @return true if this entity is a potential target for specified actor
	 */
	public boolean isTarget(Entity actor, Entity e) {
		return false;
	}

	/**
	 * is this entity a potential target for current actor
	 * 
	 * @param e
	 *            phase target
	 * @return true if this entity is a potential target for current actor
	 */
	public boolean isTarget(Entity e) {
		return false;
	}

	/**
	 * is this entity a potential actor for this phase
	 * 
	 * @param a
	 *            phase actor
	 * @return true if this entity is a potential actor
	 */
	public boolean isActor(Entity e) {
		return e.getRemainingMove() > 0
				&& !(e instanceof SummonedCreature && !((SummonedCreature) e)
						.isUnderControl());
	}

	/**
	 * proceed to phase action
	 * 
	 */
	public void action() {
		int oldCurrent = current;

		int nx, ny;
		nx = (int) clickedPoint.getX();
		ny = (int) clickedPoint.getY();
		if (getTarget() != null) {
			getActor().moveTo(getTarget(), nx, ny);
		} else {
			getActor().moveto(nx, ny);
		}

		calculatePotentialActors();

		if (current != oldCurrent || getActor().getRemainingMove() == 0) {
			BattleFrame.getInstance().clearActionPanel();
		}
		if (getActor().isUnit()) {
			SoundPlayer.getInstance().play("walk");
		} else {
			SoundPlayer.getInstance().play("indvdwalk");
		}
		getActor().clearContacts();
		getActor().calculateContacts();

		if (getActor() instanceof Unit) {
			BattleFrame.getInstance().getUnitMovementPanel().enableButtons();
		} else {
			BattleFrame.getInstance().getIndvdMovementPanel().enableButtons();
		}
		if (getActor().getOpposingContacts().size() != 0) {
			if (getActor().getRemainingMove() >= 2) {
				getActor().setRemainingmove(2);
			}
		}
		if (current == oldCurrent) { // s il ne reste qu une armee a bouger
			if (getActor().getRemainingMove() > 0) {
				// pas besoin de changer l'acteur
				setActorselection(false);
			} else {
				removePotentialActor(getActor());

				setActorselection(true);
			}
		} else {
			setActorselection(true);
		}
	}

	/**
	 * calculate phase results
	 * 
	 */
	public void calculateResults() {
		Iterator iter = Battle.getInstance().getActiveEntities().iterator();
		Entity entity;
		while (iter.hasNext()) {
			entity = (Entity) iter.next();
			if (entity.isMagic()) {
				((SpellCaster) entity).restoreMagic();
			}
			entity.setRemainingmove(0);
			if (entity.getHaste() > 0) {
				entity.consumeHaste();
			}
		}
	}

	/**
	 * determine first playing army
	 * 
	 */
	public void initCurrent() {
		if (FB.isNet()) {
			current = NetworkDX.roll(armies.length) - 1;
		} else {
			current = DX.roll(armies.length) - 1;
		}
		Battle.getInstance().setArmy(armies[current]);
		// FB.debug("initCurrent " + armies[current].getPlayerName());
		if (armies[current].isHandled()) {
			initCurrentCommand();
		}
	}

	/**
	 * @return Returns the okMove.
	 */
	public int getOkMove() {
		return okMove;
	}

	/**
	 * @return true if all commands finished their turns
	 */
	private boolean allCommandFinished() {
		int nbGroups = armies[current].getAllGroups().size();
		for (int i = 0; i < nbGroups; i++) {
			if (!hasCommandFinished(i)) {
				return false;
			}
		}
		return true;
	}

	/**
	 * test if a command has finished its movements
	 * 
	 * @param index
	 *            command index
	 * @return true if command has finished its movements
	 */
	private boolean hasCommandFinished(int index) {
		Command group = (Command) armies[current].getAllGroups().get(index);
		return doneCommands.contains(group);
	}

	/**
	 * determine first playing command
	 * 
	 */
	private void initCurrentCommand() {
		int nbGroups = armies[current].getAllGroups().size();
		currentCommandIndex = DX.roll(nbGroups) - 1;
		currentCommand = (Command) armies[current].getAllGroups().get(
				currentCommandIndex);
	}

	/**
	 * change current playing command
	 * 
	 */
	private void changeCommand() {

		int nbGroups = armies[current].getAllGroups().size();

		int i = currentCommandIndex + 1;
		if (i == nbGroups) {
			i = 0;
		}
		while (hasCommandFinished(i) && !finished()) {
			i++;
			if (i == nbGroups) {
				i = 0;
			}
		}
		currentCommandIndex = i;
		currentCommand = (Command) armies[current].getAllGroups().get(
				currentCommandIndex);
	}

}