/*
 * Copyright 2002 Farfpeuf/ AdetSoft. All rights reserved.
 * Use is subject to license terms.
 */
package fb.spells;

import javax.swing.ImageIcon;

import fb.army.entities.Entity;
import fb.army.entities.SpellCaster;
import fb.army.entities.Unit;
import fb.army.hierarchy.Army;
import fb.battlefieldelements.MagicEffect;
import fb.battlefieldelements.Warrior;
import fb.main.Battle;
import fb.util.D6;
import fb.util.DX;
import fb.util.Resources;

/**
 * Fires lightnings to the ennemy unit
 * 
 * @author farfpeuf
 * 
 */
public class Lightnings extends AttackSpell implements ValuedSpell {
	/**
	 * spell power
	 */
	private int dietoroll = 2;

	/**
	 * did target suffered any casualties ?
	 */
	private boolean targetHasCasualties = false;

	/**
	 * spell visual effects
	 */
	private static final ImageIcon[] effects = new ImageIcon[] {
			Resources.getSign("lightning0"), Resources.getSign("lightning1"),
			Resources.getSign("lightning2"), Resources.getSign("lightning3") };

	/**
	 * constructor
	 * 
	 */
	public Lightnings() {
		super();
		name = "lightnings";
	}

	/**
	 * @return string to send spell to remote computers
	 */
	protected String spellString() {
		return super.spellString() + "#" + currentValue();
	}

	/**
	 * is this entity a potential target for specified actor
	 * 
	 * @param a
	 *            spell actor
	 * @param e
	 *            spell target
	 * @return true if this entity is a potential target for specified actor
	 */
	public boolean isTarget(Entity a, Entity e) {
		return e.isUnit() && e.isEnnemy(a) && a.lineOfSightTo(e) < 9999
				&& super.isTarget(a, e);
	}

	/**
	 * is this entity a potential target for current actor
	 * 
	 * @param e
	 *            spell target
	 * @return true if this entity is a potential target for current spell
	 *         caster
	 */
	public boolean isTarget(Entity e) {
		return e.isUnit() && caster.lineOfSightTo(e) < 9999
				&& super.isTarget(e) && e.isEnnemy(caster);
	}

	/**
	 * cast the spell
	 * 
	 * @return true if the spell was successfull
	 */
	public boolean castSpell() {

		int cost = cost();
		int d6, kills = 0;
		for (int i = 0; i < dietoroll; i++) {
			d6 = D6.roll();
			if (d6 > 2) {
				kills++;
			}
		}

		if (caster.spendMagicPoints(cost)) {
			if (target.wound(kills) > 0) {
				targetHasCasualties = true;
			}
			;

			return true;
		}
		return false;
	}

	/**
	 * 
	 * @param caster
	 *            spell caster
	 * @param target
	 *            spell target
	 * @return spell cost
	 */
	public int cost(SpellCaster caster, Entity target) {
		return ((int) (0.999 + caster.stepsTo(target) / 12) + dietoroll)
				* DX.roll(3);
	}

	/**
	 * 
	 * @param caster
	 *            spell caster
	 * @param target
	 *            spell target
	 * @return spell maximum cost
	 */
	public int maxCost(SpellCaster caster, Entity target) {
		return (dietoroll + (int) (caster.stepsTo(target) / 12)) * 3;
	}

	/**
	 * 
	 * @param caster
	 *            spell caster
	 * @param target
	 *            spell target
	 * @return spell minimum cost
	 */
	public int minCost(SpellCaster caster, Entity target) {
		return dietoroll + (int) (caster.stepsTo(target) / 12);
	}

	/**
	 * 
	 * @param caster
	 *            spell caster
	 * @param target
	 *            spell target
	 * @return average cost
	 */
	public int averageCost(SpellCaster caster, Entity target) {
		return (dietoroll + (int) (caster.stepsTo(target) / 12)) * 2;
	}

	/**
	 * prepare spell visual effects
	 * 
	 */
	public void prepareEffects() {
		Warrior targetmodel;
		for (int i = 0; i < dietoroll * 5; i++) {
			targetmodel = ((Unit) target).nextTarget();
			caster.getMagics().add(
					new MagicEffect(caster, targetmodel, effects));
		}
	}

	/**
	 * end the spell
	 * 
	 * @throws InterruptedException
	 */
	public void endSpell() throws InterruptedException {

		while (target.isDying()) {
			Thread.sleep(125);
		}
		target.commitKill();
		if (target.getState() != 'a' && target.getType() != 'u') {
			if (caster.isHandled() && targetHasCasualties) {
				((Unit) target).casualtyTest();
			}
		}
		magicSynchronize();
		while (target.isWalking()) {
			Thread.sleep(125);

		}
		if (target.getState() == 'r') {
			target.commitRout();
		}
		magicSynchronize();
		Army army;
		for (int i = 0; i < Battle.getInstance().getArmies().length; i++) {
			army = Battle.getInstance().getArmies()[i];
			if (army.isHandled() && army.mustTestMorale()) {
				army.moraleTest();
			}
		}
		magicSynchronize();
		Entity entity;
		for (int i = 0; i < Battle.getInstance().getActiveEntities().size(); i++) {
			entity = (Entity) Battle.getInstance().getActiveEntities().get(i);
			// attendre la fin des mouvements
			while (entity.isWalking()) {
				Thread.sleep(125);
			}
			if (entity.getState() == 'r') {
				entity.commitRout();
			}
		}
	}

	/**
	 * @return current spell value
	 */
	public int currentValue() {
		return dietoroll;
	}

	/**
	 * @return current spell value
	 */
	public int maxValue() {
		return 30;
	}

	/**
	 * @return minimum value
	 */
	public int minValue() {
		return 2;
	}

	/**
	 * @return the value name
	 */
	public String valueName() {
		return Resources.getString("eliminationValue");
	}

	/**
	 * set the value
	 * 
	 * @param value
	 *            the new value
	 */
	public void setValue(int value) {
		dietoroll = value;
	}

	/**
	 * @param potentialTarget
	 *            the target
	 * @return weight modifiers for artificial intelligence for specified target
	 */
	public int getWeight(Entity potentialTarget) {
		int weight = super.getWeight(potentialTarget);

		if (potentialTarget.getTotalLifePoints() < 5) {
			weight += 2;
		}
		return weight;
	}

}