/*
 * Copyright 2002 Farfpeuf/ AdetSoft. All rights reserved.
 * Use is subject to license terms.
 */
package fb.util;

import java.awt.Component;
import java.awt.Container;
import java.awt.Dimension;
import java.awt.LayoutManager2;
import java.util.Enumeration;
import java.util.Hashtable;

/**
 * layout manager with percentage constraints
 * 
 * @author farfpeuf
 * 
 */

public class PercentageLayout implements LayoutManager2 {

	/**
	 * Constraint for percent layout
	 * 
	 * @author farfpeuf
	 * 
	 */
	public static class Constraint {

		/**
		 * x percentage
		 */
		private int x;

		/**
		 * y percentage
		 */
		private int y;

		/**
		 * width percentage
		 */
		private int width;

		/**
		 * height percentage
		 */
		private int height;

		/**
		 * Constructor.
		 * 
		 * @param x
		 *            x percentage
		 * @param y
		 *            y percentage
		 * @param width
		 *            width percentage
		 * @param height
		 *            height percentage
		 */
		public Constraint(int x, int y, int width, int height) {
			this.x = x;
			this.y = y;
			this.width = width;
			this.height = height;

		}
	}

	/**
	 * hashmap associating component with constraint
	 */
	private Hashtable componentConstraintsMap = new Hashtable();

	/**
	 * @see java.awt.LayoutManager2#addLayoutComponent(java.awt.Component,
	 *      java.lang.Object)
	 */
	public void addLayoutComponent(Component component, Object constraint) {
		if (constraint instanceof Constraint) {
			componentConstraintsMap.put(component, constraint);
		} else
			throw new IllegalArgumentException("Invalid constraint");
	}

	/**
	 * @see java.awt.LayoutManager#addLayoutComponent(java.lang.String,
	 *      java.awt.Component)
	 */
	public void addLayoutComponent(String constraint, Component comp) {
		throw new IllegalArgumentException("Invalid constraint");
	}

	/**
	 * @see java.awt.LayoutManager#removeLayoutComponent(java.awt.Component)
	 */
	public void removeLayoutComponent(Component component) {
		componentConstraintsMap.remove(component);
	}

	/**
	 * @see java.awt.LayoutManager#preferredLayoutSize(java.awt.Container)
	 */
	public Dimension preferredLayoutSize(Container p1) {
		int preferedSizeX = 0;
		int preferedSizeY = 0;
		Enumeration components = componentConstraintsMap.keys();
		Component component;
		Constraint constraint;
		Dimension preferedSize;
		while (components.hasMoreElements()) {
			component = (Component) components.nextElement();
			constraint = (Constraint) componentConstraintsMap.get(component);
			preferedSize = component.getPreferredSize();
			preferedSizeX += preferedSize.width * 100 / constraint.width;
			preferedSizeY += preferedSize.height * 100 / constraint.height;
		}
		return new Dimension(preferedSizeX / componentConstraintsMap.size(),
				preferedSizeY / componentConstraintsMap.size());
	}

	/**
	 * @see java.awt.LayoutManager#minimumLayoutSize(java.awt.Container)
	 */
	public Dimension minimumLayoutSize(Container p1) {
		int minx = 0;
		int miny = 0;

		Enumeration keys = componentConstraintsMap.keys();
		while (keys.hasMoreElements()) {
			Component comp = (Component) keys.nextElement();
			Constraint constraint = (Constraint) componentConstraintsMap
					.get(comp);
			Dimension min = comp.getMinimumSize();
			int mx = (int) (min.width * 100 / constraint.width);
			int my = (int) (min.height * 100 / constraint.height);
			if (mx > minx)
				minx = mx;
			if (my > miny)
				miny = my;
		}
		return new Dimension(minx, miny);
	}

	/**
	 * @see java.awt.LayoutManager2#maximumLayoutSize(java.awt.Container)
	 */
	public Dimension maximumLayoutSize(Container p1) {
		int maxx = Integer.MAX_VALUE;
		int maxy = Integer.MAX_VALUE;

		Enumeration keys = componentConstraintsMap.keys();
		while (keys.hasMoreElements()) {
			Component comp = (Component) keys.nextElement();
			Constraint constraint = (Constraint) componentConstraintsMap
					.get(comp);
			Dimension max = comp.getMaximumSize();
			int mx = max.width == Integer.MAX_VALUE ? max.width
					: (int) (max.width * 100 / constraint.width);
			int my = max.height == Integer.MAX_VALUE ? max.height
					: (int) (max.height * 100 / constraint.height);
			if (mx < maxx)
				maxx = mx;
			if (my < maxy)
				maxy = my;
		}
		return new Dimension(maxx, maxy);
	}

	/**
	 * @see java.awt.LayoutManager#layoutContainer(java.awt.Container)
	 */
	public void layoutContainer(Container p1) {
		Dimension size = p1.getSize();
		Enumeration keys = componentConstraintsMap.keys();
		while (keys.hasMoreElements()) {
			Component comp = (Component) keys.nextElement();
			Constraint constraint = (Constraint) componentConstraintsMap
					.get(comp);
			int x = (int) (size.width * constraint.x / 100);
			int y = (int) (size.height * constraint.y / 100);
			int width = (int) (size.width * constraint.width / 100);
			int height = (int) (size.height * constraint.height / 100);
			comp.setBounds(x, y, width, height);
		}
	}

	/**
	 * @see java.awt.LayoutManager2#invalidateLayout(java.awt.Container)
	 */
	public void invalidateLayout(Container p1) {
	}

	/**
	 * @see java.awt.LayoutManager2#getLayoutAlignmentY(java.awt.Container)
	 */
	public float getLayoutAlignmentY(Container p1) {
		return 0.5f;
	}

	/**
	 * @see java.awt.LayoutManager2#getLayoutAlignmentX(java.awt.Container)
	 */
	public float getLayoutAlignmentX(Container p1) {
		return 0.5f;
	}

}
